// tweakpng.cpp
//
//
/*
    Copyright (C) 1999-2004 Jason Summers

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    See the file readme.txt for more information.
*/

#include <windows.h>
#include <commdlg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <stdarg.h>
#include <commctrl.h>
#include "resource.h"
#include "tweakpng.h"
#ifdef TWPNG_HAVE_ZLIB
#include <zlib.h>
#else
#define Z_DEFAULT_COMPRESSION  (-1)
#endif

#ifdef TWPNG_SUPPORT_VIEWER
extern "C" {
	char* png_get_libpng_ver(void*);
	char* __stdcall pngdib_get_version_string(void);
}
#endif

//#define HOMEPAGE "http://pobox.com/~jason1/tweakpng/"
static const char twpng_homepage[] = TWEAKPNG_HOMEPAGE;
//#define OUR_REG_KEY "SOFTWARE\\Generic\\TweakPNG"
static const char twpng_reg_key[] = "SOFTWARE\\Generic\\TweakPNG";

#define NUMTOOLS 6

#define UPDATE_DELAY 400  // milliseconds


struct windowprefs_struct window_prefs;

static tools_t tools[NUMTOOLS];


static Png *png;
Viewer *g_viewer;
static int g_timer_set;


static HWND hwndMainList;
static HWND hwndStBar;
static int stbar_height=0;

static UINT pngchunk_cf;    // registered clipboard format

static char file_from_cmdline[MAX_PATH];
static char last_open_dir[MAX_PATH];
static char home_dir[MAX_PATH];   // dir that tweakpng.exe is in
static char orig_dir[MAX_PATH];   // current dir when program started

//int compression_level;
struct globals_struct globals;

/* table of CRCs of all 8-bit messages */
static DWORD crc_table[256];


static LRESULT CALLBACK WndProcMain(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcAbout(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcPrefs(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcSplitIDAT(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcSetSig(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK DlgProcTools(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

static void SetLVSelection(HWND hwnd, int pos1, int num);
static int OkToClosePNG(void);
static void SetTitle(void);

/* make the table for a fast crc */
void make_crc_table(void)
{
  int n;

  for (n = 0; n < 256; n++) {
    DWORD c;
    int k;

    c = (DWORD)n;
    for (k = 0; k < 8; k++)
      c = c & 1 ? 0xedb88320L ^ (c >> 1):c >> 1;

    crc_table[n] = c;
  }
}


DWORD update_crc(DWORD crc, unsigned char *buf, int len)
{
  DWORD c = crc;
  unsigned char *p = buf;
  int n = len;

  if (n > 0) do {
    c = crc_table[(c ^ (*p++)) & 0xff] ^ (c >> 8);
  } while (--n);
  return c;
}


void mesg(int severity, char *fmt, ...)
{
	va_list ap;
	static char buf[3072];
	char *t;
	int flags;

	va_start(ap, fmt);
	vsprintf(buf,fmt, ap);
	va_end(ap);

	switch(severity) {
	case MSG_S: t="Error";   flags=MB_ICONERROR;        break;
	case MSG_W: t="Warning"; flags=MB_ICONWARNING;      break;
	case MSG_I: t="Notice";  flags=MB_ICONINFORMATION;  break;
	default:    t="Error";   flags=MB_ICONWARNING;      break;  // (MSG_E)
	}
	MessageBox(globals.hwndMain,buf,t,MB_OK|flags);
}

// Some functions for reading/writing big-endian integers.

DWORD read_int32(unsigned char *x)
{
	return x[0]*256*256*256 +
	       x[1]*256*256 +
	       x[2]*256 +
	       x[3];
}

void write_int32(unsigned char *buf, DWORD x)
{
	buf[0]= (unsigned char) ((x & 0xff000000)>>24);
	buf[1]= (unsigned char) ((x & 0x00ff0000)>>16);
	buf[2]= (unsigned char) ((x & 0x0000ff00)>> 8);
	buf[3]= (unsigned char) ( x & 0x000000ff     );
}


int read_int16(unsigned char *x)
{
	return x[0]*256 +
	       x[1];
}

void write_int16(unsigned char *buf, int x)
{
	buf[0]= (unsigned char) ((x & 0x0000ff00)>> 8);
	buf[1]= (unsigned char) ( x & 0x000000ff     );
}


// return pointer to first chunk of requested type.
// return NULL if none exist
// returns the index of the chunk if index != NULL
Chunk* Png::find_first_chunk(int id, int *index)
{
	int i;
	for(i=0;i<num_chunks;i++) {
		if(chunk[i]->chunktype_id==id) {
			if(index) (*index)=i;
			return chunk[i];
		}
	}
	if(index) (*index)= -1;
	return NULL;
}

// fill in a single row (chunk) of the listview window
void Png::update_row(HWND hwnd,int i)
{
	char buf[4096];
	LV_ITEM lvi;
	int rv;

	ZeroMemory((void*)&lvi,sizeof(LV_ITEM));
	lvi.mask= LVIF_TEXT    /* LVIF_DI_SETITEM */;

	lvi.iItem=i;
	lvi.iSubItem=0;
	lvi.pszText=chunk[i]->chunktype;
	rv=ListView_SetItem(hwnd,&lvi);

	lvi.pszText=buf;

	sprintf(buf,"%u",chunk[i]->length);       // length
	lvi.iSubItem=1;
	rv=ListView_SetItem(hwnd,&lvi);

	sprintf(buf,"%08x",chunk[i]->crc);          // crc
	lvi.iSubItem=2;
	rv=ListView_SetItem(hwnd,&lvi);


	chunk[i]->get_text_descr_generic(buf);        // attributes
	lvi.iSubItem=3;
	rv=ListView_SetItem(hwnd,&lvi);

	chunk[i]->get_text_descr(buf);     // contents
	lvi.iSubItem=4;
	rv=ListView_SetItem(hwnd,&lvi);
}

// fill listview control, one line per PNG chunk
void Png::fill_listbox(HWND hwnd)
{
	int i;
	int rv;

	LV_ITEM lvi;

	ListView_DeleteAllItems(hwnd);

	ZeroMemory((void*)&lvi,sizeof(LV_ITEM));
	lvi.mask= LVIF_TEXT    /* | LVIF_DI_SETITEM ? */;


	for(i=0;i<num_chunks;i++) {
		lvi.iItem=i;
		lvi.iSubItem=0;
		lvi.pszText="";  // start with blank; update_row() will set everything
		rv=ListView_InsertItem(hwnd,&lvi);

		update_row(hwnd,i);
	}
}

// update the status bar
void Png::updatestbar(void)
{
	char buf[100];
	int i;
	DWORD s=0;
	char *type;

	if(!hwndStBar) return;

	switch(imgtype) {
	case IMG_PNG: type="PNG"; break;
	case IMG_MNG: type="MNG"; break;
	case IMG_JNG: type="JNG"; break;
	default:
		SetWindowText(hwndStBar,"No file loaded");
		goto done;
	}

	// calculate new file size
	s=8;   // the file signature
	for(i=0;i<num_chunks;i++) {
		s+= 12+chunk[i]->length;
	}

	sprintf(buf,"%s file size: %u bytes",type,s);
	SetWindowText(hwndStBar,buf);

done:
	if(g_viewer) {
		SetTimer(globals.hwndMain,1,UPDATE_DELAY,NULL);
		g_timer_set=1;
	}
}

void Png::modified(void)
{
	updatestbar();
	if(!png->dirty) {
		png->dirty=1;
		SetTitle();
	}
}

int Png::write_to_mem(unsigned char **pmem, int *plen)
{
	int s;
	int i;
	int pos;

	unsigned char *m;

	// create an image of the file in memory

	s=8;
	for(i=0;i<num_chunks;i++) {
		s+= 12+chunk[i]->length;
	}

	m=(unsigned char*)malloc(s);
	if(!m) return 0;

	CopyMemory(m,signature,8);

	pos=8;
	for(i=0;i<num_chunks;i++) {
		chunk[i]->copy_to_memory(&m[pos]);
		pos+= 12+chunk[i]->length;
	}

	*pmem = m;
	*plen = s; // should also = pos
	return 1;
}

// opens up space for new chunks and optionally creates them.
// if init is 0, opens up space for new chunks but doesn't create them
void Png::insert_chunks(int pos, int num, int init)
{
	int i;

	init_new_chunk(num_chunks+num);

	for(i=num_chunks+num-1;i>=(pos+num);i--) {
		chunk[i] = chunk[i-num];
	}

	if(init) {
		for(i=pos;i<pos+num;i++) {
			chunk[i] = new Chunk;
			chunk[i]->parentpng=this;
		}
	}

	num_chunks+=num;
}

// insert a new chunk of the specified type
void Png::new_chunk(int newid)
{
	int pos,ok,i;
	SYSTEMTIME st;
	Chunk *c;
	char newname[5];
	int plte_pos;

	pos=1;
	ok=1;

	c = new Chunk();
	c->parentpng=png;

	if(!get_name_from_id(newname,newid)) {
		mesg(MSG_S,"internal: chunk name not found for %d",newid);
		delete c;
		return;
	}
	strcpy(c->chunktype,newname);

	// plte_pos will be -1 if no PLTE chunk exists
	find_first_chunk(CHUNK_PLTE, &plte_pos);
	
	switch(newid) {
	case CHUNK_tEXt:
		pos=num_chunks-1;
		c->length=8;
		c->data = (unsigned char*)malloc(c->length);  // fixme, check for failure
		strcpy((char*)c->data,"Comment");
		break;

	case CHUNK_gAMA:
		c->length=4;
		c->data = (unsigned char*)malloc(c->length);
		write_int32(c->data,45455);
		break;

	case CHUNK_pHYs:
		c->length=9;
		c->data = (unsigned char*)malloc(c->length);
		write_int32(&c->data[0],1);
		write_int32(&c->data[4],1);
		c->data[8]=0;
		break;

	case CHUNK_sRGB:
		c->length=1;
		c->data = (unsigned char*)malloc(c->length);
		c->data[0]=0;
		break;

	case CHUNK_cHRM:
		c->length=32;
		c->data = (unsigned char*)malloc(c->length);
		write_int32(&c->data[ 0],31270);
		write_int32(&c->data[ 4],32900);
		write_int32(&c->data[ 8],64000);
		write_int32(&c->data[12],33000);
		write_int32(&c->data[16],30000);
		write_int32(&c->data[20],60000);
		write_int32(&c->data[24],15000);
		write_int32(&c->data[28], 6000);
		break;

	case CHUNK_bKGD:
		if(plte_pos>=0) pos=plte_pos+1;
		switch(colortype) {
		case 3: c->length=1; break;
		case 0: case 4: c->length=2; break;
		case 2: case 6: c->length=6; break;
		default: return;
		}
		c->data = (unsigned char*)calloc(c->length,1);
		break;

	case CHUNK_sBIT:
		switch(colortype) {
		case 0: c->length=1; break;
		case 2: case 3: c->length=3; break;
		case 4: c->length=2; break;
		case 6: c->length=4; break;
		default: return;
		}
		c->data = (unsigned char*)calloc(c->length,1);
		for(i=0;i<(int)c->length;i++) {
			c->data[i]= (colortype==3)?8:bitdepth;
		}
		break;

	case CHUNK_PLTE:
		if(colortype!=2 && colortype!=3 && colortype!=6) {
			mesg(MSG_E,"Palette is not allowed for grayscale images");
			return;
		}
		c->length=3;
		c->data = (unsigned char*)calloc(c->length,1);
		break;

	case CHUNK_tRNS:
		if(plte_pos>=0) pos=plte_pos+1;
		switch(colortype) {
		case 3: c->length=1; break;
		case 0: c->length=2; break;
		case 2: c->length=6; break;
		case 4: case 6:
			mesg(MSG_E,"Transparency chunk is not allowed for this image, since "
				"it already has a full alpha channel");
			return;
		default: return;
		}
		c->data = (unsigned char*)calloc(c->length,1);
		if(colortype==3) c->data[0]=255;  // default to opaque
		break;

	case CHUNK_IHDR:
		pos=0;
		c->length=13;
		c->data = (unsigned char*)calloc(c->length,1);
		write_int32(&c->data[ 0],1); // width
		write_int32(&c->data[ 4],1); // height
		c->data[8]=1; // bit depth
		break;

	case CHUNK_IEND:
		pos=num_chunks;
		c->length=0;
		c->data = NULL;
		break;

	case CHUNK_tIME:
		pos=num_chunks-1;
		c->length=7;
		c->data = (unsigned char*)malloc(c->length);

		GetSystemTime(&st);
		write_int16(&c->data[0],(int)st.wYear);
		c->data[2]= (unsigned char)st.wMonth;
		c->data[3]= (unsigned char)st.wDay;
		c->data[4]= (unsigned char)st.wHour;
		c->data[5]= (unsigned char)st.wMinute;
		c->data[6]= (unsigned char)st.wSecond;
		break;

	default:
		ok=0;
	}

	if(!ok) {
		mesg(MSG_W,"Not implemented");
		return;
	}

	if(pos<0) pos=0;
	if(pos>num_chunks) pos=num_chunks;

	insert_chunks(pos,1,1);

	chunk[pos]=c;
	c->after_init();
	c->chunkmodified();

	fill_listbox(hwndMainList);
	SetLVSelection(hwndMainList,pos,1);
	modified();
}


void Png::edit_chunk(int n)
{
	int r,i;

	if(n<0 || n>=num_chunks) return;
	r= chunk[n]->edit();

	if(r>0) {
		if(r==2) {
			// update all
			for(i=0;i<num_chunks;i++) {
				update_row(hwndMainList,i);
			}
		}
		else if(r==1) {
			update_row(hwndMainList,n);
		}
		else {
			mesg(MSG_S,"internal error: change code %d",r);
		}
		modified();
	}
}

// save to disk
// returns 1 on success, 0 on failure
int Png::write_file(char *fn)
{
	int i;
	HCURSOR hcur;
	DWORD written;
	HANDLE fh;

	fh=CreateFile(fn,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,
		FILE_ATTRIBUTE_NORMAL,NULL);
	if(fh==INVALID_HANDLE_VALUE) {
		mesg(MSG_E,"Can't write file (%s)",fn);
		return 0;
	}

	hcur=SetCursor(LoadCursor(NULL,IDC_WAIT));

	WriteFile(fh,(LPVOID)signature,8,&written,NULL);

	for(i=0;i<num_chunks;i++) {
		chunk[i]->write_to_file(fh,0);
	}
	CloseHandle(fh);
	SetCursor(hcur);
	return 1;
}

// make sure we have room for have n chunks in the chunk[] array
void Png::init_new_chunk(int n)   
{
	if(n<chunks_alloc) return;
	// oops -- overran our chunks array. try to alloc a bigger one
	chunks_alloc += 200;
	chunk=(Chunk**)realloc((void*)chunk, chunks_alloc*sizeof(Chunk*));
	if(!chunk) {
		mesg(MSG_S,"can't alloc memory for chunks array");
		num_chunks=0;
	}
}

void Png::delete_chunk(int n)
{
	int i;
	if(n<0 || n>=num_chunks) return;

	delete chunk[n];
	for(i=n+1;i<num_chunks;i++) {
		chunk[i-1]=chunk[i];
	}
	num_chunks--;
}

void Png::move_chunk(int n, int delta)  // move chunk n by delta
{
	Chunk *tmpchunk;
	int moveto;
	int i;

	if(n<0 || n>=num_chunks) return;
	moveto=n+delta;

	if(moveto<0) moveto=0;
	if(moveto>=num_chunks) moveto=num_chunks-1;

	if(moveto==n) return;

	if(moveto>n) {        // move down
		for(i=0;i<(moveto-n);i++) {
			tmpchunk=chunk[n+i];       //swap(n+i,n+i+1)
			chunk[n+i]=chunk[n+i+1];
			chunk[n+i+1]=tmpchunk;
		}
	}
	else {        // move up      (moveto < n)
		for(i=(n-moveto-1);i>=0;i--) {
			tmpchunk=chunk[moveto+i];   //swap(moveto+i+1,moveto+i)
			chunk[moveto+i]=chunk[moveto+i+1];
			chunk[moveto+i+1]=tmpchunk;
		}
	}
}


static unsigned char sig_png[] = {137,80,78,71,13,10,26,10};
static unsigned char sig_mng[] = {138,77,78,71,13,10,26,10};
static unsigned char sig_jng[] = {139,74,78,71,13,10,26,10};

void Png::set_signature(void)
{
	switch(imgtype) {
	case IMG_PNG: memcpy(signature,sig_png,8); break;
	case IMG_MNG: memcpy(signature,sig_mng,8); break;
	case IMG_JNG: memcpy(signature,sig_jng,8); break;
	}
}

int Png::read_signature(HANDLE fh)
{
	DWORD n;
	int r;

	r=ReadFile(fh,(LPVOID)signature,8,&n,NULL);
	if(!r || n<8) return 0;
	if(!memcmp(signature,sig_png,8)) return IMG_PNG;
	if(!memcmp(signature,sig_mng,8)) return IMG_MNG;
	if(!memcmp(signature,sig_jng,8)) return IMG_JNG;
	return IMG_UNKNOWN;
}

int Png::read_next_chunk(HANDLE fh)
{
	DWORD n;
	Chunk *c;
	unsigned char fbuf[8];
	DWORD ccrc;
	int r;

	// first 4 bytes are the chunk data length,
	// next 4 bytes are the chunk type
	r=ReadFile(fh,(LPVOID)fbuf,8,&n,NULL);
	if(!r || n==0) {  // this is normal; we've reached the end of file
		return 0;
	}

	if(n!=8) {
		mesg(MSG_W,"Garbage found at end of file");
		return 0;
	}

	// allocate a Chunk structure for this new chunk
	c = new Chunk();

	c->parentpng = this;  // chunks sometimes depend other chunks, ...

	c->length= read_int32(&fbuf[0]);

	memcpy(c->chunktype,&fbuf[4],4);
	c->chunktype[4]='\0';  // null-terminate for convenience

	// now read the data
	if(c->length>0) {

		// A sanity test for the chunk length. This ought to compare to the 
		// remaining bytes instead of the total size, but it's no big deal.
		if(c->length > pngfilesize) {
			mesg(MSG_W,"Bogus chunk length found. This may indicate garbage at the end of the file.");
			delete c;
			return 0;
		}

		c->data = (unsigned char*)malloc(c->length);
		if(!c->data) {
			mesg(MSG_S,"Can't allocate memory for chunk");
			delete c;
			return 0;
		}
	}

	r=ReadFile(fh,(LPVOID)c->data,c->length,&n,NULL);
	if(!r || n!=c->length) {
		mesg(MSG_W,"Garbage found at end of file");
		delete c;
		return 0;
	}

	// now read the CRC
	r=ReadFile(fh,(LPVOID)fbuf,4,&n,NULL);
	if(!r || n!=4) {
		mesg(MSG_W,"Garbage found at end of file");
		delete c;
		return 0;
	}

	c->crc = read_int32(&fbuf[0]);

	// check the crc
	ccrc=c->calc_crc();

	if(c->crc != ccrc) {
		mesg(MSG_W,"Incorrect crc for %s chunk (is %08x, should be %08x)",
			c->chunktype, c->crc, ccrc);
		c->crc=ccrc;  // correct it
	}

	init_new_chunk(num_chunks);  // make sure chunks array is large enough
	chunk[num_chunks++]=c;

	c->after_init();
	return 1;
}

Png::Png(void)
{
	num_chunks=0;
	chunk=NULL;
	chunks_alloc=0;
	strcpy(filename,"untitled");
	named=0;
	dirty=0;

	chunk=(Chunk**)malloc(200*sizeof(Chunk*));
	chunks_alloc=200;
	imgtype=IMG_PNG;

	width=1;
	height=1;
	bitdepth=1;
	colortype=0;
	compression=0;
	compressionfilter=0;
	interlace=0;

	memcpy(signature,sig_png,8);

	valid=1;
}


Png::Png(char *fn)
{
	int okay;
	HANDLE fh;
	valid=0;

	num_chunks=0;
	chunk=NULL;
	chunks_alloc=0;
	strcpy(filename,fn);
	named=1;
	dirty=0;

	chunk=(Chunk**)malloc(200*sizeof(Chunk*));
	chunks_alloc=200;

	colortype=255;  // random invalid value

	fh=CreateFile(fn,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,NULL);
	if(fh==INVALID_HANDLE_VALUE) {
		mesg(MSG_E,"Can't open file (%s)",fn);
		return;
	}

	pngfilesize=GetFileSize(fh,NULL);  // note, this will fail for files over 4 GB

	imgtype=read_signature(fh);
	okay=(imgtype>=1);

	if(!okay) {
		SetForegroundWindow(globals.hwndMain);
		mesg(MSG_E,"Unrecognized file format\n\nThis is not a valid PNG file.");
		updatestbar();
		return;
	}

	while(okay) {
		okay=read_next_chunk(fh);
	}

	CloseHandle(fh);
	updatestbar();
	valid=1;
}


Png::~Png(void)
{
	int i;

	// free individual chunks
	for(i=0;i<num_chunks;i++) {
		if(chunk[i]) delete chunk[i];
	}

	// free chunk list
	if(chunk) free(chunk);
}


///////////////////////////////////////////////////////////////////

static int RegisterClasses(void)
{
	WNDCLASS  wc;

	wc.style = CS_DBLCLKS|CS_HREDRAW;
	wc.lpfnWndProc = (WNDPROC)WndProcMain;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = globals.hInst;
	wc.hIcon = LoadIcon(globals.hInst,"ICONMAIN");
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)GetStockObject(DKGRAY_BRUSH);
	wc.lpszMenuName =  "MENUMAIN";
	wc.lpszClassName = "TWEAKPNGMAIN";
	if(!RegisterClass(&wc)) {
		return 0;
	}

	wc.style= CS_HREDRAW|CS_VREDRAW;
	wc.lpfnWndProc = (WNDPROC)WndProcEditPal;
	wc.hIcon = NULL;
	wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName =  NULL;
	wc.lpszClassName = "TWEAKPNGEDITPAL";
	if(!RegisterClass(&wc)) {
		return 0;
	}

	return 1;
}


int StoreWindowPos(HWND hwnd, struct windowpos_struct *q)
{
	WINDOWPLACEMENT wp;

	if(!IsWindow(hwnd)) return 0;
	ZeroMemory(&wp,sizeof(wp));
	wp.length=sizeof(wp);
	if(!GetWindowPlacement(hwnd,&wp)) return 0;
	q->x= wp.rcNormalPosition.left;
	q->y= wp.rcNormalPosition.top;
	q->w= wp.rcNormalPosition.right - q->x;
	q->h= wp.rcNormalPosition.bottom - q->y;
	q->max= (wp.showCmd == SW_SHOWMAXIMIZED);  // this isn't used
	return 1;
}


static int SaveSettings(void)
{
	HKEY key;
	LONG r;
	DWORD disp, tmpd;
	char v[24];
	int i;

	r=RegCreateKeyEx(HKEY_CURRENT_USER,twpng_reg_key,0,NULL,
		REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&key,&disp);
	if(r!=ERROR_SUCCESS) {
		mesg(MSG_E,"Cannot save settings to registry");
		return 0;
	}

	for(i=0;i<NUMTOOLS;i++) {
		sprintf(v,"tool%dn",i);
		r=RegSetValueEx(key,v,0,REG_SZ,(LPBYTE)tools[i].name  ,1+strlen(tools[i].name));

		sprintf(v,"tool%dc",i);
		r=RegSetValueEx(key,v,0,REG_SZ,(LPBYTE)tools[i].cmd   ,1+strlen(tools[i].cmd));

		sprintf(v,"tool%dp",i);
		r=RegSetValueEx(key,v,0,REG_SZ,(LPBYTE)tools[i].params,1+strlen(tools[i].params));
	}

	r=RegSetValueEx(key,"deflate_lvl",0,REG_DWORD,(LPBYTE)&globals.compression_level,sizeof(globals.compression_level));
	r=RegSetValueEx(key,"show_viewer",0,REG_DWORD,(LPBYTE)&globals.autoopen_viewer,sizeof(DWORD));
	r=RegSetValueEx(key,"use_gamma",0,REG_DWORD,(LPBYTE)&globals.use_gamma,sizeof(DWORD));
	r=RegSetValueEx(key,"use_custombg",0,REG_DWORD,(LPBYTE)&globals.use_custombg,sizeof(DWORD));
	r=RegSetValueEx(key,"use_imagebg",0,REG_DWORD,(LPBYTE)&globals.use_imagebg,sizeof(DWORD));
	r=RegSetValueEx(key,"windowbg",0,REG_DWORD,(LPBYTE)&globals.window_bgcolor,sizeof(DWORD));
	r=RegSetValueEx(key,"zoom",0,REG_DWORD,(LPBYTE)&globals.vsize,sizeof(DWORD));

	if(IsWindow(hwndMainList)) {
		for(i=0;i<5;i++) {
			window_prefs.column_width[i]=ListView_GetColumnWidth(hwndMainList,i);
		}
	}

	StoreWindowPos(globals.hwndMain,&window_prefs.main);

	RegSetValueEx(key,"windowpos",0,REG_BINARY,(LPBYTE)(&window_prefs),sizeof(window_prefs));

	tmpd = (DWORD)globals.custombgcolor;
	RegSetValueEx(key,"bgcolor",0,REG_DWORD,(LPBYTE)&tmpd,sizeof(DWORD));
	RegSetValueEx(key,"custcolors",0,REG_BINARY,(LPBYTE)&globals.custcolors,16*sizeof(COLORREF));

	RegCloseKey(key);

	return 1;
}

static void ReadSettings(void)
{
	HKEY key;
	LONG r;
	DWORD datasize,tmpd;
	int i;
	char v[24];
	//prefs_t tmp_prefs;

	globals.compression_level=Z_DEFAULT_COMPRESSION;
	globals.use_gamma=1;
	globals.custombgcolor = RGB(128,128,128);
	globals.use_custombg = 1;
	globals.use_imagebg = 1;
	for(i=0;i<16;i++) globals.custcolors[i] = RGB(0,0,0);
	globals.autoopen_viewer=0;
	globals.window_bgcolor=TWPNG_WBG_SAMEASIMAGE;

	for(i=0;i<NUMTOOLS;i++) {
		strcpy(tools[i].name,"");
		strcpy(tools[i].cmd,"");
		strcpy(tools[i].params,"");
	}
	strcpy(tools[0].name,"Default viewer");
	window_prefs.version = 1;
	window_prefs.column_width[0]=50;
	window_prefs.column_width[1]=60;
	window_prefs.column_width[2]=60;
	window_prefs.column_width[3]=130;
	window_prefs.column_width[4]=600;

	window_prefs.main.x=0;
	window_prefs.main.y=0;
	window_prefs.main.w=600;
	window_prefs.main.h=300;
	window_prefs.main.max=0;

	window_prefs.text.x=0;
	window_prefs.text.y=0;
	window_prefs.text.w=500;
	window_prefs.text.h=300;
	window_prefs.text.max=0;

	window_prefs.plte.x=0;
	window_prefs.plte.y=0;
	window_prefs.plte.w=490;
	window_prefs.plte.h=440;
	window_prefs.plte.max=0;

	window_prefs.viewer.x=0;
	window_prefs.viewer.y=100;
	window_prefs.viewer.w=350;
	window_prefs.viewer.h=250;
	window_prefs.viewer.max=0;

	if(RegOpenKeyEx(HKEY_CURRENT_USER,twpng_reg_key,0,KEY_READ,&key)
		!= ERROR_SUCCESS) return;

	for(i=0;i<NUMTOOLS;i++) {
		sprintf(v,"tool%dn",i); datasize=MAX_TOOL_NAME-1;
		r=RegQueryValueEx(key,v,NULL,NULL,(LPBYTE)tools[i].name,&datasize);
		tools[i].name[MAX_TOOL_NAME-1]='\0';

		sprintf(v,"tool%dc",i); datasize=MAX_TOOL_CMD-1;
		r=RegQueryValueEx(key,v,NULL,NULL,(LPBYTE)tools[i].cmd,&datasize);
		tools[i].cmd[MAX_TOOL_CMD-1]='\0';

		sprintf(v,"tool%dp",i); datasize=MAX_TOOL_PARAMS-1;
		r=RegQueryValueEx(key,v,NULL,NULL,(LPBYTE)tools[i].params,&datasize);
		tools[i].params[MAX_TOOL_PARAMS-1]='\0';
	}

	datasize=sizeof(globals.compression_level);
	r=RegQueryValueEx(key,"deflate_lvl",NULL,NULL,(LPBYTE)&globals.compression_level,&datasize);

	datasize=sizeof(window_prefs);
	r=RegQueryValueEx(key,"windowpos",NULL,NULL,(LPBYTE)(&window_prefs),&datasize);

	datasize=sizeof(DWORD);
	r=RegQueryValueEx(key,"show_viewer",NULL,NULL,(LPBYTE)(&globals.autoopen_viewer),&datasize);

	datasize=sizeof(DWORD);
	r=RegQueryValueEx(key,"use_gamma",NULL,NULL,(LPBYTE)(&globals.use_gamma),&datasize);
	datasize=sizeof(DWORD);
	r=RegQueryValueEx(key,"use_custombg",NULL,NULL,(LPBYTE)(&globals.use_custombg),&datasize);
	datasize=sizeof(DWORD);
	r=RegQueryValueEx(key,"use_imagebg",NULL,NULL,(LPBYTE)(&globals.use_imagebg),&datasize);
	datasize=sizeof(DWORD);
	r=RegQueryValueEx(key,"windowbg",NULL,NULL,(LPBYTE)(&globals.window_bgcolor),&datasize);
	datasize=sizeof(DWORD);
	r=RegQueryValueEx(key,"zoom",NULL,NULL,(LPBYTE)(&globals.vsize),&datasize);

	datasize=sizeof(DWORD);
	r=RegQueryValueEx(key,"bgcolor",NULL,NULL,(LPBYTE)&tmpd,&datasize);
	if(r==ERROR_SUCCESS) {
		globals.custombgcolor = (COLORREF)tmpd;
	}

	datasize = 16*sizeof(COLORREF);
	r=RegQueryValueEx(key,"custcolors",NULL,NULL,(LPBYTE)(&globals.custcolors),&datasize);
	RegCloseKey(key);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
						  int nCmdShow)
{
	MSG msg;
	char buf[MAX_PATH];
	HACCEL hAccTable;
	int p;

	ZeroMemory(&globals,sizeof(struct globals_struct));

	globals.hInst=hInstance;
	globals.hwndMain=NULL;
	hwndMainList=NULL;
	hwndStBar=NULL;
	png=NULL;
	g_viewer=NULL;
	g_timer_set=0;
	globals.vborder=4;
	globals.vsize=TWPNG_VS_FIT;

	InitCommonControls();

	pngchunk_cf = RegisterClipboardFormat("pngchunks");

	if(!RegisterClasses()) return 0;
#ifdef TWPNG_SUPPORT_VIEWER
	Viewer::GlobalInit();
#endif
	globals.hcurDrag = LoadCursor(globals.hInst,"DRAGCURSOR");
	globals.hcurDrag2 = LoadCursor(globals.hInst,"DRAGCURSOR2");

	strcpy(last_open_dir,"");

	strncpy(buf,lpCmdLine,MAX_PATH-1); buf[MAX_PATH-1]='\0';
	if(buf[0]=='"') {    // if quoted, strip quotes
		strcpy(file_from_cmdline,&buf[1]);
		file_from_cmdline[strlen(file_from_cmdline)-1]='\0';
	}
	else {
		strcpy(file_from_cmdline,buf);
	}

	make_crc_table();

	strcpy(orig_dir,"");
	strcpy(home_dir,"");

	GetCurrentDirectory(MAX_PATH,orig_dir);

	// record the directory we were loaded from
	if(GetModuleFileName(globals.hInst,home_dir,MAX_PATH)) {
		// strip filename, leaving directory path
		p=strlen(home_dir)-1;
		while(p>=0) {
			if(home_dir[p]=='\\') {
				home_dir[p]='\0';
				break;
			}
			p--;
		}
	}

	ReadSettings();

	globals.hwndMain = CreateWindow(
		"TWEAKPNGMAIN","TweakPNG",
		WS_VISIBLE|WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN,
		CW_USEDEFAULT,CW_USEDEFAULT,window_prefs.main.w,window_prefs.main.h,
		NULL,NULL,globals.hInst,NULL);
	if (!globals.hwndMain) return 0;

	DragAcceptFiles(globals.hwndMain,TRUE);
	hAccTable=LoadAccelerators(globals.hInst,"ACCELTABLE");

	while(GetMessage(&msg,NULL,0,0)){
		if (!TranslateAccelerator(globals.hwndMain, hAccTable, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	return (msg.wParam);
}

// set the titlebar text for the main window
static void SetTitle(void)
{
	char buf[1024];
	if(png) {
		sprintf(buf,"%s%s - TweakPNG",png->filename,png->dirty?"*":"");
		SetWindowText(globals.hwndMain,buf);
	}
	else {
		SetWindowText(globals.hwndMain,"TweakPNG");
	}
}


static void NewPng(void)
{
	if(png) {
		delete png;
		png=NULL;
	}
	ListView_DeleteAllItems(hwndMainList);
	png=new Png();
	png->updatestbar();
	SetTitle();
}

// handles loading a new png file
static int OpenPngByName(char *fn)
{
	if(png) {
		delete png;
		png=NULL;
	}
	ListView_DeleteAllItems(hwndMainList);
	png=new Png(fn);
	png->fill_listbox(hwndMainList);

	if(!png->valid) {
		delete png;
		png=NULL;
		SetTitle();
		return 0;
	}
	SetTitle();

	return 1;
}

static int OpenPngFromMenu(HWND hwnd)
{
	char fn[MAX_PATH];
	OPENFILENAME ofn;

	strcpy(fn,"");

	ZeroMemory(&ofn,sizeof(OPENFILENAME));

	ofn.lStructSize=sizeof(OPENFILENAME);
	ofn.hwndOwner=hwnd;
	ofn.hInstance=NULL;
	ofn.lpstrFilter="PNG, MNG, JNG\0*.png;*.mng;*.jng\0PNG\0*.png\0MNG\0*.mng\0JNG\0*.jng\0All files\0*.*\0\0";
	ofn.nFilterIndex=1;
	ofn.lpstrTitle="Open image file";
	if(strlen(last_open_dir))
		ofn.lpstrInitialDir=last_open_dir;  // else NULL ==> current dir
	ofn.lpstrFile=fn;
	ofn.nMaxFile=MAX_PATH;
	ofn.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY|OFN_NOCHANGEDIR;

	if(!GetOpenFileName(&ofn)) {
		return 0;  // user canceled
	}

	strcpy(last_open_dir,fn);
	last_open_dir[ofn.nFileOffset]='\0'; // chop off filename; save the path for next time

	return OpenPngByName(fn);
}


void DroppedFiles(HDROP hDrop)
{
	UINT rv;
	char fn[MAX_PATH];


	rv=DragQueryFile(hDrop,0xFFFFFFFF,NULL,0);
 
	if(rv==1) {  	// check number of files dropped
		rv=DragQueryFile(hDrop,0,fn,MAX_PATH);
		if(rv) {
			if(OkToClosePNG()) OpenPngByName(fn);
		}
	}
	else {
		mesg(MSG_E,"Cannot accept multiple dropped files");
	}
	DragFinish(hDrop);
}

// Do a few validity checks on the number and ordering of chunks
// sets pointer to string describing the problem. Setting msg to NULL
// makes this function display a dialog box about any errors (or even
// if all is okay)
// This function only checks the overall structure, not parameter values,
// etc.
//
// msgmode==0: prints results (good or bad), returns 1 if okay, 0 if errors found
// msgmode==1: if errors found, asks if you want to save, returns 0 if user says no, otherwise 1
int Png::check_validity(int msgmode)
{
	char *m;
	int i;
	int e=0;   // error count
	int t;
	int prev_chunk;
	char buf[500];

	int idat_seen=0, plte_seen=0, chrm_seen=0, gama_seen=0;
	int iccp_seen=0, sbit_seen=0, srgb_seen=0, bkgd_seen=0;
	int hist_seen=0, trns_seen=0, phys_seen=0, time_seen=0;

	m="No problems found.";

	if(imgtype!=IMG_PNG) {
		if(msgmode==0) {
			e++; m="Can only check PNG files";
		}
		goto done;
	}

	i=0;
	prev_chunk=CHUNK_UNKNOWN;

	while(i<num_chunks && !e) {
		t=chunk[i]->chunktype_id;   // to save typing

		if(chunk[i]->is_critical() && t!=CHUNK_IHDR && t!=CHUNK_PLTE
			&& t!=CHUNK_IDAT && t!=CHUNK_IEND)
		{
			e++; m="Unrecognized critical chunk";
		}

		if(i==0 && t!=CHUNK_IHDR) {
			e++; m="First chunk must be IHDR";
		}
		else if((i==num_chunks-1) && t!=CHUNK_IEND) {
			e++; m="Last chunk must be IEND";
		}
		else {
			switch(t) {
			case CHUNK_IHDR:
				if(i!=0) {
					e++; m="Misplaced or extra IHDR";
				}
				break;
			case CHUNK_IEND:
				if(i!=num_chunks-1) {
					e++; m="Misplaced or extra IEND";
				}
				break;
			case CHUNK_IDAT:
				if(idat_seen && prev_chunk!=CHUNK_IDAT) {
					e++; m="IDAT chunks must be consecutive";
				}
				idat_seen++;
				break;
			case CHUNK_PLTE:
				if(plte_seen) { e++; m="Multiple PLTE chunks not allowed"; }
				else if(colortype==0 || colortype==4) {
					e++; m="PLTE chunk not allowed in grayscale image";
				}
				else if(idat_seen) { e++; m="PLTE must appear before IDAT"; }
				else if(bkgd_seen) { e++; m="bKGD must appear after PLTE"; }
				else if(hist_seen) { e++; m="hIST must appear after PLTE"; }
				else if(trns_seen) { e++; m="tRNS must appear after PLTE"; }
				plte_seen++;
				break;
			case CHUNK_tIME:
				if(time_seen) { e++; m="Multiple tIME chunks not allowed"; }
				time_seen++;
				break;
			case CHUNK_cHRM:
				if(chrm_seen) { e++; m="Multiple cHRM chunks not allowed"; }
				else if(plte_seen) { e++; m="cHRM must appear before PLTE"; }
				else if(idat_seen) { e++; m="cHRM must appear before IDAT"; }
				chrm_seen++;
				break;
			case CHUNK_gAMA:
				if(gama_seen) { e++; m="Multiple gAMA chunks not allowed"; }
				else if(plte_seen) { e++; m="gAMA must appear before PLTE"; }
				else if(idat_seen) { e++; m="gAMA must appear before IDAT"; }
				gama_seen++;
				break;
			case CHUNK_iCCP:
				if(iccp_seen) { e++; m="Multiple iCCP chunks not allowed"; }
				else if(plte_seen) { e++; m="iCCP must appear before PLTE"; }
				else if(idat_seen) { e++; m="iCCP must appear before IDAT"; }
				iccp_seen++;
				break;
			case CHUNK_sBIT:
				if(sbit_seen) { e++; m="Multiple sBIT chunks not allowed"; }
				else if(plte_seen) { e++; m="sBIT must appear before PLTE"; }
				else if(idat_seen) { e++; m="sBIT must appear before IDAT"; }
				sbit_seen++;
				break;
			case CHUNK_sRGB:
				if(srgb_seen) { e++; m="Multiple sRGB chunks not allowed"; }
				else if(plte_seen) { e++; m="sRGB must appear before PLTE"; }
				else if(idat_seen) { e++; m="sRGB must appear before IDAT"; }
				srgb_seen++;
				break;
			case CHUNK_bKGD:
				if(bkgd_seen) { e++; m="Multiple bKGD chunks not allowed"; }
				else if(idat_seen) { e++; m="sRGB must appear before IDAT"; }
				bkgd_seen++;
				break;
			case CHUNK_hIST:
				if(hist_seen) { e++; m="Multiple hIST chunks not allowed"; }
				else if(idat_seen) { e++; m="hIST must appear before IDAT"; }
				hist_seen++;
				break;
			case CHUNK_tRNS:
				if(trns_seen) { e++; m="Multiple tRNS chunks not allowed"; }
				else if(idat_seen) { e++; m="tRNS must appear before IDAT"; }
				trns_seen++;
				break;
			case CHUNK_pHYs:
				if(phys_seen) { e++; m="Multiple pHYs chunks not allowed"; }
				else if(idat_seen) { e++; m="pHYs must appear before IDAT"; }
				phys_seen++;
				break;
			case CHUNK_sPLT:
				if(idat_seen) { e++; m="sPLT must appear before IDAT"; }
				break;

			}
		}

		prev_chunk=t;
		i++;
	}

	if(!e) {  // if no errors yet, we'll test some more things
		if(num_chunks<1) {
			e++; m="No chunks. Not valid.";
		}
		else if(!idat_seen) {
			e++; m="Required IDAT chunk not found";
		}
		else if(colortype==3 && !plte_seen) {
			e++; m="Required PLTE chunk not found";
		}
	}

done:
	if(msgmode==0) {
		MessageBox(globals.hwndMain,m,"Validity check",MB_OK|(e?MB_ICONWARNING:MB_ICONINFORMATION));
		return (!e);
	}
	//else

	if(e) {
		sprintf(buf,"A problem was detected with the current file:\n\n%s\n\n"
			"Do you want to save it anyway?",m);
		if(MessageBox(globals.hwndMain,buf,"Validity check",MB_OKCANCEL|MB_ICONWARNING)!=IDOK)
		{
			return 0;
		}
	}
	return 1;

}

// returns 1 if saved, else 0
static int SavePngAs(HWND hwnd)
{
	OPENFILENAME ofn;
	char fn[MAX_PATH];

	if(!png->check_validity(1)) return 0;

	strcpy(fn,png->filename);

	ZeroMemory(&ofn,sizeof(OPENFILENAME));
	ofn.lStructSize=sizeof(OPENFILENAME);
	ofn.hwndOwner=hwnd;

	if(png->imgtype==IMG_MNG) {
		ofn.lpstrFilter="MNG (*.mng)\0*.mng\0\0";
		ofn.lpstrDefExt="mng";
	}
	else if(png->imgtype==IMG_JNG) {
		ofn.lpstrFilter="JNG (*.jng)\0*.jng\0\0";
		ofn.lpstrDefExt="jng";
	}
	else {
		ofn.lpstrFilter="PNG (*.png)\0*.png\0\0";
		ofn.lpstrDefExt="png";
	}

	ofn.nFilterIndex=1;
	ofn.lpstrTitle="Save Image As...";
	ofn.lpstrFile=fn;
	ofn.nMaxFile=MAX_PATH;
	ofn.Flags=OFN_PATHMUSTEXIST|OFN_HIDEREADONLY|OFN_OVERWRITEPROMPT|OFN_NOCHANGEDIR;

	if(GetSaveFileName(&ofn)) {
		if(png->write_file(ofn.lpstrFile)) {
			strcpy(png->filename,ofn.lpstrFile);
			png->named=1;
			png->dirty=0;
			SetTitle();
			return 1;
		}
	}
	return 0;
}

// returns 1 if saved, else 0
static int SavePng(HWND hwnd)
{
	if(png->named) {
		if(!png->check_validity(1)) return 0;
		if(png->write_file(png->filename)) {
			png->dirty=0;
			SetTitle();
			return 1;
		}
		return 0;
	}
	else {
		return SavePngAs(hwnd);
	}
}

static void DeleteChunks(void)          // delete all selected items
{
	int i;
	int numdeleted=0;
	int firstdeleted;

	firstdeleted= png->num_chunks;

	for(i=png->num_chunks-1;i>=0;i--) {
		if(ListView_GetItemState(hwndMainList,i,LVIS_SELECTED) & LVIS_SELECTED) {
			png->delete_chunk(i);
			numdeleted++;
			firstdeleted=i;
		}
	}

	if(numdeleted>0) {
		png->fill_listbox(hwndMainList);
	}

	if(firstdeleted<0) firstdeleted=0;
	if(firstdeleted>(png->num_chunks-1)) firstdeleted=png->num_chunks-1;
	SetLVSelection(hwndMainList,firstdeleted,1);
	png->modified();
}

// selects a range of items, and sets the focus to the first of that range
static void SetLVSelection(HWND hwnd, int pos1, int num)
{
	int i;

	for(i=pos1;i<pos1+num;i++) {
		ListView_SetItemState(hwnd,i,
			LVIS_SELECTED| ((i==pos1)?LVIS_FOCUSED:0),
			LVIS_SELECTED| ((i==pos1)?LVIS_FOCUSED:0)  );
	}
	ListView_EnsureVisible(hwnd,pos1,FALSE);
}


static void LVUnselectAll(HWND hwnd)
{
	int i,c;

	c=ListView_GetItemCount(hwnd);
	for(i=0;i<c;i++) {
		ListView_SetItemState(hwnd,i,0,LVIS_SELECTED);
	}
}

static int GetLVFocus(HWND hwnd)
{
	int i;
	for(i=0;i<png->num_chunks;i++) {
		if(ListView_GetItemState(hwnd,i,LVIS_FOCUSED) & LVIS_FOCUSED) {
			return i;
		}
	}
	return 0;
}	

/* returns the selected item if exactly one item is selected,
 * otherwise returns -1 */
static int GetLVSelection(HWND hwnd)
{
	int i;

	if(ListView_GetSelectedCount(hwnd)!=1) return(-1);

	for(i=0;i<png->num_chunks;i++) {
		if(ListView_GetItemState(hwnd,i,LVIS_SELECTED) & LVIS_SELECTED) {
			return i;
		}
	}
	return(-1);
}

static void PasteChunks(void)
{
	DWORD msize=0;
	HGLOBAL hClip;
	unsigned char* lpClip;
	DWORD p;
	int r,inspos1,inspos,numnewchunks;

	inspos1=inspos=GetLVFocus(hwndMainList);
	numnewchunks=0;

	if(!IsClipboardFormatAvailable(pngchunk_cf)) return;

	OpenClipboard(NULL);
	hClip=GetClipboardData(pngchunk_cf);
	if(hClip) {
		lpClip= (unsigned char*)GlobalLock(hClip);
		if(lpClip) {
			//msize=GlobalSize(hClip);
			msize=read_int32(&lpClip[0]);
			p=4;
			while(p<msize) {
				png->insert_chunks(inspos,1,1);  // this is wrong, I should make a new chunk, then insert it if things go okay
				r= png->chunk[inspos]->init_from_memory(&lpClip[p], msize-p);
				if(!r) break;  // error occurred
				inspos++;
				numnewchunks++;
				p+=r;
			}
			GlobalUnlock(hClip);
			png->fill_listbox(hwndMainList);

			// reselect the new or changed chunks
			SetLVSelection(hwndMainList,inspos1,numnewchunks);
			png->modified();
		}
	}
	CloseClipboard();
}

// returns 1 if any chunks were selected, and successfully copied
// otherwize 0
static int CopyChunks(void)
{
	int i;
	DWORD msize=0;
	HGLOBAL hClip;
	unsigned char* lpClip;
	int p;

	for(i=0;i<png->num_chunks;i++) {
		if(ListView_GetItemState(hwndMainList,i,LVIS_SELECTED) & LVIS_SELECTED) {
			msize+= 12+png->chunk[i]->length;
		}
	}
	if(msize<1) return 0;  // nothing selected
	msize+=4;

	hClip=GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE,msize);
	lpClip= (unsigned char*)GlobalLock(hClip);

	// I don't know how to figure out the size of the clipboard data
	// when retrieving it. Am I just stupid?
	// GlobalSize() often returns a larger than actual size ...
	// As a workaround, the first 4 bytes of the clipboard data
	// will contain the total length (including those 4 bytes)

	write_int32(&lpClip[0],msize);
	p=4;
	for(i=0;i<png->num_chunks;i++) {
		if(ListView_GetItemState(hwndMainList,i,LVIS_SELECTED) & LVIS_SELECTED) {
			png->chunk[i]->copy_to_memory((unsigned char*)&lpClip[p]);
			p+= 12+png->chunk[i]->length;
		}
	}
	GlobalUnlock(hClip);

	OpenClipboard(NULL);
	EmptyClipboard();
	SetClipboardData(pngchunk_cf,hClip);
	CloseClipboard();
	return 1;
}

static void CutChunks(void)
{
	if(CopyChunks()) {
		DeleteChunks();
	}
}

static void ImportChunk(void)
{
	OPENFILENAME ofn;
	char fn[MAX_PATH];
	HANDLE fh;
	Chunk *c;
	int pos;
	DWORD n;

	pos=GetLVFocus(hwndMainList);

	strcpy(fn,"");
	ZeroMemory(&ofn,sizeof(OPENFILENAME));

	ofn.lStructSize=sizeof(OPENFILENAME);
	ofn.hwndOwner=globals.hwndMain;
	ofn.hInstance=NULL;
	ofn.lpstrFilter="*.chunk\0*.chunk\0*.*\0*.*\0\0";
	ofn.nFilterIndex=1;
	ofn.lpstrTitle="Import chunk...";
//	if(strlen(last_open_dir))
//		ofn.lpstrInitialDir=last_open_dir;  // else NULL ==> current dir
	ofn.lpstrFile=fn;
	ofn.nMaxFile=MAX_PATH;
	ofn.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY|OFN_NOCHANGEDIR;

	if(!GetOpenFileName(&ofn)) return;

	fh=CreateFile(fn,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,NULL);
	if(fh==INVALID_HANDLE_VALUE) {
		mesg(MSG_E,"Can't open file (%s)",fn);
		return;
	}

	c=new Chunk;
	c->length=GetFileSize(fh,NULL)-4;
	ReadFile(fh,(LPVOID)c->chunktype,4,&n,NULL);
	c->chunktype[4]='\0';
	c->data=(unsigned char*)malloc(c->length);
	ReadFile(fh,(LPVOID)c->data,c->length,&n,NULL);
	CloseHandle(fh);

	c->parentpng=png;
	c->after_init();
	c->chunkmodified();


	png->insert_chunks(pos,1,0);
	png->chunk[pos]=c;

	png->fill_listbox(hwndMainList);
	png->modified();
	SetLVSelection(hwndMainList,pos,1);
}

static void ExportChunk(void)
{
	OPENFILENAME ofn;
	char fn[MAX_PATH];
	HANDLE fh;
	int n;

	n=GetLVSelection(hwndMainList);
	if(n<0) {
		mesg(MSG_I,"You must select a single chunk first");
		return;
	}

	strcpy(fn,"");

	ZeroMemory(&ofn,sizeof(OPENFILENAME));
	ofn.lStructSize=sizeof(OPENFILENAME);
	ofn.hwndOwner=globals.hwndMain;
	ofn.lpstrFilter="*.chunk\0*.chunk\0*.*\0*.*\0\0";
	ofn.nFilterIndex=1;
	ofn.lpstrTitle="Export chunk to file...";
	ofn.lpstrFile=fn;
	ofn.nMaxFile=MAX_PATH;
	ofn.Flags=OFN_PATHMUSTEXIST|OFN_HIDEREADONLY|OFN_OVERWRITEPROMPT|OFN_NOCHANGEDIR;
	ofn.lpstrDefExt="dat";

	if(GetSaveFileName(&ofn)) {
		fh=CreateFile(ofn.lpstrFile,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,
			FILE_ATTRIBUTE_NORMAL,NULL);
		if(fh==INVALID_HANDLE_VALUE) {
			mesg(MSG_E,"Can't create file");
		}
		else {
			png->chunk[n]->write_to_file(fh,1);
			CloseHandle(fh);
		}
	}
}

static void MoveChunkUp(void)
{
	int i;

	if(ListView_GetSelectedCount(hwndMainList)<1) return;

	// mark selected chunks so we can reselect them later
	for(i=0;i<png->num_chunks;i++) {
		png->chunk[i]->flag= (ListView_GetItemState(hwndMainList,i,LVIS_SELECTED) & LVIS_SELECTED)?1:0;
	}

	// can't move up if first is selected
	if(png->chunk[0]->flag) return;

	for(i=1;i<png->num_chunks;i++) {
		if(png->chunk[i]->flag) {
			png->move_chunk(i,-1);
		}
	}
		
	png->fill_listbox(hwndMainList);

	for(i=png->num_chunks-1;i>=0;i--) {   // reselect moved chunks
		if(png->chunk[i]->flag) SetLVSelection(hwndMainList,i,1);
	}

	png->modified();
}


static void MoveChunkDown(void)
{
	int i;

	if(ListView_GetSelectedCount(hwndMainList)<1) return;

	// mark selected chunks so we can reselect them later
	for(i=0;i<png->num_chunks;i++) {
		png->chunk[i]->flag= (ListView_GetItemState(hwndMainList,i,LVIS_SELECTED) & LVIS_SELECTED)?1:0;
	}

	// can't move down if last is selected
	if(png->chunk[png->num_chunks-1]->flag) return;

	for(i=png->num_chunks-1;i>=0;i--) {
		if(png->chunk[i]->flag) {
			png->move_chunk(i,1);
		}
	}
		
	png->fill_listbox(hwndMainList);

	for(i=0;i<png->num_chunks;i++) {   // reselect moved chunks
		if(png->chunk[i]->flag) SetLVSelection(hwndMainList,i,1);
	}

	png->modified();
}


// split chunk n at size ssize (and repeat if repeat==1)
int Png::split_idat(int n, int ssize, int repeat)
{
	Chunk *c;
	int new_chunks;
	int i;
	int bytes_used;
	int thissize;
	char buf[200];

	c=chunk[n];

	// how many chunks will there be after the split
	if(!repeat) {
		new_chunks=2;
	}
	else {
		new_chunks= (c->length + ssize-1) / ssize;
	}

	if(new_chunks>100) {
		sprintf(buf,"This will create %d new chunks. Are you sure you want to continue?",new_chunks);
		i=MessageBox(globals.hwndMain,buf,"Warning",MB_ICONQUESTION|MB_YESNO|MB_DEFBUTTON2);
		if(i!=IDYES) return 0;
	}

	chunk[n]=new Chunk;

	insert_chunks(n,new_chunks-1,1);  // -1 because we start with one already
	for(i=n;i<n+new_chunks;i++) {
		chunk[i]->parentpng=png;
		strcpy(chunk[i]->chunktype,c->chunktype);
	}

	bytes_used=0;

	for(i=n;i<n+new_chunks;i++) {
		if(i<n+new_chunks-1) {
			thissize=ssize;
		}
		else {
			thissize= c->length - bytes_used;  // == all remaining bytes
		}

		chunk[i]->length = thissize;
		if(thissize>0) {
			chunk[i]->data = (unsigned char*)malloc(thissize);
			memcpy((void*)chunk[i]->data,(void*)&c->data[bytes_used],thissize);
		}
		else {
			chunk[i]->data = NULL;
		}

		bytes_used += thissize;
	}

	delete c;

	for(i=n;i<n+new_chunks;i++) {
		chunk[i]->after_init();
		chunk[i]->chunkmodified();
	}
	modified();
	fill_listbox(hwndMainList);

	// reselect the new or changed chunks
	SetLVSelection(hwndMainList,n,new_chunks);

	return 1;
}

static void SplitIDAT(void)
{
	int n,id;
	
	n=GetLVSelection(hwndMainList);
	if(n<0) {
		mesg(MSG_I,"Select a single IDAT chunk first");
		return;
	}

	id=png->chunk[n]->chunktype_id;

	if(id!=CHUNK_IDAT && id!=CHUNK_JDAT) {
		mesg(MSG_I,"Select an IDAT chunk first");
		return;
	}

	png->chunk[n]->index=n;
	DialogBoxParam(globals.hInst,"DLG_SPLITIDAT",globals.hwndMain,(DLGPROC)DlgProcSplitIDAT,
		(LPARAM)png->chunk[n]);

}

// mode  0: combine selected  1: combine all
static void CombineIDAT(int mode)
{
	int i, first, last;
	DWORD len,pos;
	Chunk *c;
	unsigned char *newdata;

	first= -1; last= -1;

	if(mode==1) {
		for(i=0;i<png->num_chunks;i++) {
			if( (png->chunk[i]->chunktype_id == CHUNK_IDAT) ||
				(png->chunk[i]->chunktype_id == CHUNK_JDAT))
			{
				if(first == -1) first=i;
			}
			else {
				if(first!= -1 && last == -1) last= i-1;
			}
		}
	}
	else {

		for(i=0;i<png->num_chunks;i++) {
			if(ListView_GetItemState(hwndMainList,i,LVIS_SELECTED) & LVIS_SELECTED) {
				if( (png->chunk[i]->chunktype_id != CHUNK_IDAT) &&
					(png->chunk[i]->chunktype_id != CHUNK_JDAT))
				{
					mesg(MSG_E,"Only IDAT chunks may be selected");
					return;
				}
				if(first == -1) {
					first = i;
				}
				else if(last != -1) {
					mesg(MSG_E,"Selected chunks must be consecutive");
					return;
				}
			}
			else {    // not selected
				if(first != -1) {
					if(last == -1) {
						last = i-1;
					}
				}
			}
		}
	}
	if(first == -1) {
		if(mode==1)
			mesg(MSG_E,"No IDAT chunks present");
		else
			mesg(MSG_E,"Must select IDAT chunks first");
		return;
	}
	if(last== -1) last=png->num_chunks-1;  // not really legal, since IDAT can't be last, but...
	if(last<=first) {    // only one chunk?
		if(mode==1) {
			LVUnselectAll(hwndMainList);
			SetLVSelection(hwndMainList,first,1);
		}
		else {
			mesg(MSG_E,"Must select more than one IDAT chunk");
		}
		return;
	}

	// calculate total length of data in new IDAT chunk
	len=0;
	for(i=first;i<=last;i++) {
		len+=png->chunk[i]->length;
	}

	newdata= (unsigned char*)malloc(len);
	if(!newdata) {
		mesg(MSG_S,"can't allocate memory");
		return;
	}

	pos=0;
	for(i=first;i<=last;i++) {
		if(png->chunk[i]->length>0) {
			memcpy((void*)&newdata[pos],
				(void*)png->chunk[i]->data,png->chunk[i]->length);
			pos+=png->chunk[i]->length;
		}
	}

	c=new Chunk;
	c->parentpng = png;
	c->data=newdata;
	c->length=len;
	strcpy(c->chunktype,png->chunk[first]->chunktype);
	c->after_init();

	c->chunkmodified();    // set crc

	for(i=first;i<=last;i++) {
		delete png->chunk[i];
	}

	png->chunk[first]=c;

	// we're deleting  (last-first) chunks
	for(i=first+1;i<=png->num_chunks-(last-first+1);i++) {
		png->chunk[i] = png->chunk[i+(last-first)];
	}

	png->num_chunks -= (last-first);
	
	png->fill_listbox(hwndMainList);
	SetLVSelection(hwndMainList,first,1);

	png->modified();
}

static void DblClickOnList(void)
{
	int s;

	s=GetLVSelection(hwndMainList);
	if(s>=0 && s<png->num_chunks) {
		png->edit_chunk(s);
	}
}

static int CreateMainWindows(HWND hwnd)
{
	LV_COLUMN lvc;
	RECT r;
	DWORD style;

	hwndStBar=CreateStatusWindow(WS_CHILD|WS_VISIBLE,"",hwnd,ID_STBAR);
	if(!hwndStBar) {
		mesg(MSG_S,"Cannot create status bar");
		return 0;
	}

	GetWindowRect(hwndStBar,&r);
	stbar_height=r.bottom-r.top;

	GetClientRect(hwnd,&r);

	hwndMainList=CreateWindowEx(WS_EX_CLIENTEDGE,
		WC_LISTVIEW, "main listview",
		WS_VISIBLE|WS_CHILD|LVS_REPORT|LVS_SHOWSELALWAYS|LVS_NOSORTHEADER,
		r.left, r.top, r.right-r.left, r.bottom-r.top-stbar_height,
		hwnd,NULL,globals.hInst,NULL);
	if(!hwndMainList) {
		mesg(MSG_S,"Cannot create listview control");
		return 0;
	}

	style=ListView_GetExtendedListViewStyle(hwndMainList);
	ListView_SetExtendedListViewStyle(hwndMainList,style|LVS_EX_FULLROWSELECT);

	lvc.mask= LVCF_FMT | LVCF_TEXT | LVCF_WIDTH;
    lvc.cchTextMax=0;
	lvc.iSubItem=0;

	lvc.pszText="Chunk";
	lvc.cx= window_prefs.column_width[0];
	lvc.fmt= LVCFMT_LEFT;
	ListView_InsertColumn(hwndMainList,0,&lvc);

	lvc.pszText="Length";
	lvc.cx= window_prefs.column_width[1];
	lvc.fmt= LVCFMT_RIGHT;
	ListView_InsertColumn(hwndMainList,1,&lvc);

	lvc.pszText="CRC";
	lvc.cx= window_prefs.column_width[2];
	lvc.fmt= LVCFMT_LEFT;
	ListView_InsertColumn(hwndMainList,2,&lvc);

	lvc.pszText="Attributes";
	lvc.cx= window_prefs.column_width[3];
	lvc.fmt= LVCFMT_LEFT;
	ListView_InsertColumn(hwndMainList,3,&lvc);

	lvc.pszText="Contents";
	lvc.cx= window_prefs.column_width[4];
	lvc.fmt= LVCFMT_LEFT;
	ListView_InsertColumn(hwndMainList,4,&lvc);

	return 1;
}

// n is the tool number (0-5)
static void RunTool(int n)
{
	char tmpname[MAX_PATH+2];
	char buf[1000];
	HINSTANCE h;

	if(n<0 || n>=NUMTOOLS) return;
	if(!strlen(tools[n].name)) {
		mesg(MSG_E,"Tool %d not defined",n+1);
		return;
	}

	SetCurrentDirectory(home_dir);

	if(!GetTempPath(MAX_PATH,tmpname)) {
		//mesg(MSG_S,"Can't find a temp directory");
		//return;
		strcpy(tmpname,"");
	}

	// make sure pathname ends in a backslash or is empty
	if(strlen(tmpname)) {
		if(tmpname[strlen(tmpname)-1] != '\\') strcat(tmpname,"\\");
	}

	if(png->imgtype==IMG_MNG) {
		strcat(tmpname,"temp$$$$.mng");
	}
	else if(png->imgtype==IMG_JNG) {
		strcat(tmpname,"temp$$$$.jng");
	}
	else {
		strcat(tmpname,"temp$$$$.png");
	}

	if(!png->write_file(tmpname)) {
		goto done;    // uh oh, the goto police are gonna get me...
	}

	if(strlen(tools[n].cmd)) {
		sprintf(buf,"%s%s %s",tools[n].params,strlen(tools[n].params)?" ":"",tmpname);
		h=ShellExecute(globals.hwndMain,"open",tools[n].cmd,buf,NULL,SW_SHOWNORMAL);
	}
	else {
		h=ShellExecute(globals.hwndMain,"open",tmpname,NULL,NULL,SW_SHOWNORMAL);
	}

	if( (UINT)h <= 32) {
		mesg(MSG_E,"Cannot execute tool (error %u)",(UINT)h);
	}

done:
	SetCurrentDirectory(orig_dir);
}

// hwnd is the window to attach the menu to, x,y is the mouse position
static void ContextMenu(HWND hwnd, int x, int y)
{
	int sel;
	int cmd;
	HMENU menu;

	sel=ListView_GetSelectedCount(hwndMainList);
	if(sel<1) return;

	menu=CreatePopupMenu();

	AppendMenu(menu,(sel==1)?MF_ENABLED:MF_GRAYED,ID_EDITCHUNK,"&Edit Chunk...");
	AppendMenu(menu,MF_ENABLED,ID_DELCHUNK,"&Delete");
	AppendMenu(menu,MF_ENABLED,ID_MOVEUP,"Move &Up");
	AppendMenu(menu,MF_ENABLED,ID_MOVEDOWN,"Mo&ve Down");
	AppendMenu(menu,MF_SEPARATOR,0,NULL);
	AppendMenu(menu,MF_ENABLED,ID_CUT,"Cu&t");
	AppendMenu(menu,MF_ENABLED,ID_COPY,"&Copy");
	AppendMenu(menu,MF_ENABLED,ID_PASTE,"&Paste");

	cmd=TrackPopupMenuEx(menu, TPM_LEFTALIGN|TPM_TOPALIGN|TPM_NONOTIFY|TPM_RETURNCMD|
		TPM_RIGHTBUTTON,x,y,hwnd,NULL);
	DestroyMenu(menu);

	switch(cmd) {
	case ID_EDITCHUNK:    DblClickOnList();  return;
	case ID_DELCHUNK:     DeleteChunks();    return;
	case ID_MOVEUP:       MoveChunkUp();     return;
	case ID_MOVEDOWN:     MoveChunkDown();   return;
	case ID_COPY:   CopyChunks();    return;
	case ID_CUT:    CutChunks();     return;
	case ID_PASTE:  PasteChunks();   return;
	}
}

static int OkToClosePNG(void)
{
	char buf[500];
	int x,ret;

	if(!png) return 1;

	if(png->dirty) {
		SetForegroundWindow(globals.hwndMain);
		wsprintf(buf,"Save changes to %s?",png->filename);
		x=MessageBox(globals.hwndMain,buf,"TweakPNG",MB_YESNOCANCEL|MB_ICONWARNING|MB_DEFBUTTON3);
		if(x==IDYES) {
			ret=SavePng(globals.hwndMain);
			return ret?1:0;
		}
		else if(x==IDNO) {
			return 1;
		}
		else if(x==IDCANCEL) {
			return 0;
		}
		return 0;
	}
	else {
		return 1; // not modified
	}
}

void update_viewer()
{
#ifdef TWPNG_SUPPORT_VIEWER

	unsigned char *m;
	int len;
	int rv;

	if(!g_viewer) return;

	if(!png) {
		g_viewer->Update(NULL,0,0);
		return;
	}

	rv= png->write_to_mem(&m,&len);
	if(!rv) {
		g_viewer->Update(NULL,0,0);
		return;
	}

	g_viewer->Update(m,len,png->imgtype);
	free(m);
#endif
}

static LRESULT CALLBACK WndProcMain(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id;
	RECT r;
	int i;
	char buf[150];

	id=LOWORD(wParam);

	switch(msg) {
	case WM_ERASEBKGND:
		return 1;

	case WM_CREATE:
		globals.hwndMain=hwnd;

		// create the listview control, statusbar, etc.
		if(!CreateMainWindows(hwnd)) return -1;  // abort program
			
		if(strlen(file_from_cmdline)) {
			OpenPngByName(file_from_cmdline);
		}
		if(globals.autoopen_viewer) {
			PostMessage(hwnd,WM_COMMAND,ID_IMGVIEWER,(LPARAM)0);
		}
		return 0;

	case WM_CLOSE:
		// close requested e.g. by user clicking on 'x' icon.
		if(OkToClosePNG()) break;
		return 0; // handle this message --> will not be closed

	case WM_DESTROY:
		if(g_timer_set) {
			KillTimer(hwnd,1);
			g_timer_set=0;
		}
		SaveSettings();
		if(png) delete png;
		png=NULL;
		PostQuitMessage(0);
		return 0;

	case WM_SETFOCUS:
		if(hwndMainList) {
			SetFocus(hwndMainList);
			return 0;
		}
		break;

	case WM_CONTEXTMENU:
		ContextMenu(hwnd,LOWORD(lParam),HIWORD(lParam));
		return 0;

	case WM_SIZE:
		GetClientRect(hwnd,&r);
		SetWindowPos(hwndStBar,NULL,  // this is probably not the right way to do this...
			r.bottom-stbar_height,r.left,r.right,r.bottom,
			SWP_NOZORDER);
		SetWindowPos(hwndMainList,NULL,
			r.left,r.top,r.right-r.left,r.bottom-r.top-stbar_height,
			SWP_NOZORDER);
		return 0;

	case WM_DROPFILES:
		DroppedFiles((HDROP)wParam);
		return 0;

	case WM_TIMER:
		if(g_timer_set) {
			KillTimer(hwnd,1);
			g_timer_set=0;
		}
		update_viewer();
		return 0;

	case WM_INITMENU:
		{
			HMENU m, mtools;
			UINT x;
			int sel;

			// commands requiring an open file
			static UINT cmdlist1[] = {ID_SAVE,ID_SAVEAS,
				ID_NEWTEXT,ID_NEWBKGD,ID_NEWGAMA,ID_NEWIEND,ID_NEWPHYS,
				ID_NEWIHDR,
				ID_NEWSRGB,ID_NEWTIME,ID_NEWCHRM,ID_NEWTRNS,ID_NEWSBIT,
				ID_NEWPLTE,ID_COMBINEALLIDAT,
				ID_IMPORTCHUNK,ID_SIGNATURE,ID_CHECKPNG,
				ID_TOOL_1,ID_TOOL_2,ID_TOOL_3,ID_TOOL_4,ID_TOOL_5,ID_TOOL_6,
				0};
			// commands requiring exactly 1 selected chunk
			static UINT cmdlist2[] = {ID_EDITCHUNK,
				ID_SPLITIDAT,ID_EXPORTCHUNK,
				0};
			// requiring 1 or more selected chunks
			static UINT cmdlist3[] = {ID_DELCHUNK,ID_COPY,ID_CUT,ID_MOVEUP,ID_MOVEDOWN,
				0};

			static UINT toolsi[] = {ID_TOOL_1,ID_TOOL_2,ID_TOOL_3,ID_TOOL_4,
				ID_TOOL_5,ID_TOOL_6};

			m=(HMENU)wParam;

			mtools=CreatePopupMenu();
#ifdef TWPNG_SUPPORT_VIEWER
			AppendMenu(mtools,MF_STRING|MF_ENABLED,ID_IMGVIEWER,"&Show Image Viewer\tF7");
			AppendMenu(mtools,MF_SEPARATOR,0,NULL);
#endif
			for(i=0;i<NUMTOOLS;i++) {
				if(strlen(tools[i].name)) {
					sprintf(buf,"&%d  %s\tCtrl+%d",i+1,tools[i].name,i+1);
					AppendMenu(mtools,MF_STRING|MF_ENABLED,toolsi[i],buf);
				}
			}

			// the Tools menu has an index of 5
			ModifyMenu(m,4,MF_POPUP|MF_BYPOSITION|MF_ENABLED,(UINT)mtools,"&Tools");

			sel=0;
			if(png) {
				if(hwndMainList) {
					sel=ListView_GetSelectedCount(hwndMainList);
				}
			}

			x= MF_BYCOMMAND | (png?MF_ENABLED:MF_GRAYED);
			for(i=0;cmdlist1[i];i++) {
				EnableMenuItem(m,cmdlist1[i],x);
			}

			x= MF_BYCOMMAND | ( (sel==1)?MF_ENABLED:MF_GRAYED );
			for(i=0;cmdlist2[i];i++) {
				EnableMenuItem(m,cmdlist2[i],x);
			}

			x= MF_BYCOMMAND | ( (sel>=1)?MF_ENABLED:MF_GRAYED );
			for(i=0;cmdlist3[i];i++) {
				EnableMenuItem(m,cmdlist3[i],x);
			}

			// commands with unusual rules for enabling
			EnableMenuItem(m,ID_COMBINEIDAT, MF_BYCOMMAND |
				((sel>1)?MF_ENABLED:MF_GRAYED) );
			
			x= MF_BYCOMMAND | (png?MF_ENABLED:MF_GRAYED);
			EnableMenuItem(m,ID_PASTE,MF_BYCOMMAND |
				(IsClipboardFormatAvailable(pngchunk_cf))?x:MF_GRAYED);

			CheckMenuItem(m,ID_IMGVIEWER,MF_BYCOMMAND|
				(g_viewer?MF_CHECKED:MF_UNCHECKED));
			return 0;
		}

	case WM_NOTIFY:
		{
			LPNMHDR  lpnmh = (LPNMHDR) lParam;
			if(!png) break;
			switch(lpnmh->code) {
			case NM_DBLCLK:
			case NM_RETURN:
				if(lpnmh->hwndFrom==hwndMainList) {
					DblClickOnList();
					return 0;
				}
			}
			break;
		}

	case WM_COMMAND:
		// commands always available
		switch(id) {
#ifdef TWPNG_SUPPORT_VIEWER
		case ID_IMGVIEWER:
			if(g_viewer) {
				g_viewer->Close();
				if(g_viewer) { delete g_viewer; g_viewer=NULL; }
				globals.autoopen_viewer=0;
			}
			else {
				g_viewer = new Viewer(globals.hwndMain);
				update_viewer();
				globals.autoopen_viewer=1;
			}
			return 0;
		case ID_UPDATEVIEWER:
			update_viewer();
			return 0;
#endif
		case ID_EXIT:
			if(OkToClosePNG()) DestroyWindow(hwnd);
			return 0;
		case ID_OPEN:
			if(OkToClosePNG()) OpenPngFromMenu(hwnd);
			return 0;
		case ID_NEW:   
			if(OkToClosePNG()) NewPng();
			return 0;

		case ID_PREFS:
			DialogBox(globals.hInst,"DLG_PREFS",hwnd,(DLGPROC)DlgProcPrefs);
			return 0;

		case ID_EDITTOOLS:
			DialogBox(globals.hInst,"DLG_TOOLS",globals.hwndMain,(DLGPROC)DlgProcTools);
			return 0;

		case ID_ABOUT:
			DialogBox(globals.hInst,"DLG_ABOUT",hwnd,(DLGPROC)DlgProcAbout);
			return 0;

		case ID_HELPCONTENTS:
			SetCurrentDirectory(home_dir);
			if((UINT)ShellExecute(globals.hwndMain,"open","tweakpng.txt",
				NULL,NULL,SW_SHOWNORMAL)<=32)
			{
				mesg(MSG_E,"Cannot display help file");
			}
			SetCurrentDirectory(orig_dir);
			return 0;
		case ID_SWITCHWINDOW:
			if(g_viewer) {
				if(GetFocus()==g_viewer->hwndViewer) {
					SetFocus(globals.hwndMain);
				}
				else {
					SetFocus(g_viewer->hwndViewer);
				}
			}
			return 0;
		}

		// commands only available if we have an image loaded
		if(png) switch(id) {
		case ID_SIGNATURE:
			DialogBoxParam(globals.hInst,"DLG_SIGNATURE",globals.hwndMain,(DLGPROC)DlgProcSetSig,
				(LPARAM)png);
			return 0;

		case ID_SAVE:         SavePng(hwnd);     return 0;
		case ID_SAVEAS:       SavePngAs(hwnd);   return 0;
		case ID_CHECKPNG:     png->check_validity(0);  return 0;

		case ID_EDITCHUNK:    DblClickOnList();  return 0;
		case ID_DELCHUNK:     DeleteChunks();    return 0;
		case ID_MOVEUP:       MoveChunkUp();     return 0;
		case ID_MOVEDOWN:     MoveChunkDown();   return 0;

		case ID_NEWBKGD: png->new_chunk(CHUNK_bKGD); return 0;
		case ID_NEWCHRM: png->new_chunk(CHUNK_cHRM); return 0;
		case ID_NEWGAMA: png->new_chunk(CHUNK_gAMA); return 0;
		case ID_NEWIEND: png->new_chunk(CHUNK_IEND); return 0;
		case ID_NEWIHDR: png->new_chunk(CHUNK_IHDR); return 0;
		case ID_NEWPHYS: png->new_chunk(CHUNK_pHYs); return 0;
		case ID_NEWPLTE: png->new_chunk(CHUNK_PLTE); return 0;
		case ID_NEWSBIT: png->new_chunk(CHUNK_sBIT); return 0;
		case ID_NEWSRGB: png->new_chunk(CHUNK_sRGB); return 0;
		case ID_NEWTEXT: png->new_chunk(CHUNK_tEXt); return 0;
		case ID_NEWTIME: png->new_chunk(CHUNK_tIME); return 0;
		case ID_NEWTRNS: png->new_chunk(CHUNK_tRNS); return 0;

		case ID_CUT:    CutChunks();     return 0;
		case ID_COPY:   CopyChunks();    return 0;
		case ID_PASTE:  PasteChunks();   return 0;
		case ID_SELECTALL:
			SetLVSelection(hwndMainList,0,png->num_chunks);
			return 0;

		case ID_SPLITIDAT:    SplitIDAT();       return 0;
		case ID_COMBINEIDAT:     CombineIDAT(0);     return 0;
		case ID_COMBINEALLIDAT:  CombineIDAT(1);     return 0;

		case ID_IMPORTCHUNK:  ImportChunk();     return 0;
		case ID_EXPORTCHUNK:  ExportChunk();     return 0;

		case ID_TOOL_1: RunTool(0); return 0;
		case ID_TOOL_2: RunTool(1); return 0;
		case ID_TOOL_3: RunTool(2); return 0;
		case ID_TOOL_4: RunTool(3); return 0;
		case ID_TOOL_5: RunTool(4); return 0;
		case ID_TOOL_6: RunTool(5); return 0;

		}

	}
	return (DefWindowProc(hwnd, msg, wParam, lParam));
}


static BOOL CALLBACK DlgProcAbout(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	char buf[4000],buf1[1000];
	WORD id;
	id=LOWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:

#ifdef _DEBUG
#define TWPNG_DEBUGSTR " (Debug build)"
#else
#define TWPNG_DEBUGSTR ""
#endif
		wsprintf(buf,"TweakPNG, version %s\r\nPNG image file maniplution utility\r\n"
			"Copyright (c) %s by Jason Summers\r\n%s%s\r\nHome page: %s\r\n",
			TWEAKPNG_VER_STRING,TWEAKPNG_COPYRIGHT_DATE,__DATE__,TWPNG_DEBUGSTR,twpng_homepage);
		strcat(buf,"\r\nThis program is distributed under the terms "
			"of the GNU General Public License. Please read the file tweakpng.txt for more information.\r\n");

#ifdef TWPNG_HAVE_ZLIB
		sprintf(buf1,"\r\nUses zlib, version %s",ZLIB_VERSION);
		strcat(buf,buf1);
#else
		strcat(buf,"\r\nCompiled without compression/decompression support.");
#endif

#ifdef TWPNG_SUPPORT_VIEWER
		wsprintf(buf1,"\r\nUses libpng, version %s.",png_get_libpng_ver(NULL));
		lstrcat(buf,buf1);

		wsprintf(buf1,"\r\nUses pngdib, version %s.",pngdib_get_version_string());
		lstrcat(buf,buf1);
#else
		lstrcat(buf,"\r\nCompiled without image viewing support.");
#endif
		SetDlgItemText(hwnd,IDC_ABOUTTEXT,buf);
		return 1;
	case WM_COMMAND:
		switch(id) {
		case IDOK:
		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
	}
	return 0;
}

// Returns the "name" of files with a given extension.
// format = ".png", ".jng", ".mng"
// returns zero if does not exist
// caller supplies buf[500]
static int get_reg_name_for_ext(const char *format, char *logicalname)
{
	HKEY key1;
	LONG rv;
	int retval;
	DWORD datatype,datasize;

	retval=0;
	key1=NULL;
	sprintf(logicalname,"");

	rv=RegOpenKeyEx(HKEY_CLASSES_ROOT,format,0,KEY_READ,&key1);
	if(rv!=ERROR_SUCCESS) goto abort;

	datasize=500;
	rv=RegQueryValueEx(key1,NULL,NULL,&datatype,(BYTE*)logicalname,&datasize);
	if(rv!=ERROR_SUCCESS) goto abort;
	if(datatype!=REG_SZ) goto abort;
	if(datasize<2) goto abort;
	RegCloseKey(key1); key1=NULL;
	retval=1;

abort:
	if(key1) RegCloseKey(key1);
	return retval;
}

static int is_explorer_menu_set(void)
{
	HKEY key1;
	LONG rv;
	int retval;
	char logicalname[500];
	char buf[500];
	int x;


	retval=0;
	key1=NULL;

	x=get_reg_name_for_ext(".png",logicalname);
	if(!x) goto abort;

	_snprintf((char*)buf,500,"%s\\shell\\tweakpng",logicalname);
	// we only want to know if this key exists
	rv=RegOpenKeyEx(HKEY_CLASSES_ROOT,buf,0,KEY_READ,&key1);

	if(rv!=ERROR_SUCCESS) goto abort;

	retval=1;

abort:
	if(key1) RegCloseKey(key1);
	return retval;
}

static char *fmts[] = { ".png",".mng",".jng" };
static char *dflt_fmt_name[] = { "png_auto_file","mng_auto_file","jng_auto_file" };

// Add to Windows Explorer right-click menu for .[pmj]ng files.
static void add_to_explorer_menu(void)
{
	int i;
	int x;
	char logicalname[500];
	HKEY key1;
	DWORD disp;
	LONG rv;
	char buf[500];
	char cmd[500];

	for(i=0;i<=2;i++) {
		x=get_reg_name_for_ext(fmts[i],logicalname);
		if(!x) {
			sprintf(logicalname,dflt_fmt_name[i]);
			// create HKCR\\.*ng if necessary
			rv=RegCreateKeyEx(HKEY_CLASSES_ROOT,fmts[i],0,NULL,
				REG_OPTION_NON_VOLATILE,KEY_WRITE,NULL,&key1,&disp);
			if(rv==ERROR_SUCCESS) {
				RegSetValueEx(key1,"",0,REG_SZ,(BYTE*)logicalname,1+strlen(logicalname));
				RegCloseKey(key1);
			}
		}

		// create HKCR\\[png_auto_file] if necessary
		rv=RegCreateKeyEx(HKEY_CLASSES_ROOT,logicalname,0,NULL,
			REG_OPTION_NON_VOLATILE,KEY_WRITE,NULL,&key1,&disp);
		if(rv==ERROR_SUCCESS) {
			RegCloseKey(key1);
		}

		// create HKCR\\[png_auto_file]\shell
		sprintf(buf,"%s\\shell",logicalname);
		rv=RegCreateKeyEx(HKEY_CLASSES_ROOT,buf,0,NULL,
			REG_OPTION_NON_VOLATILE,KEY_WRITE,NULL,&key1,&disp);
		if(rv==ERROR_SUCCESS) {
			RegCloseKey(key1);
		}

		// create HKCR\\[png_auto_file]\shell\tweakpng
		sprintf(buf,"%s\\shell\\tweakpng",logicalname);
		rv=RegCreateKeyEx(HKEY_CLASSES_ROOT,buf,0,NULL,
			REG_OPTION_NON_VOLATILE,KEY_WRITE,NULL,&key1,&disp);
		if(rv==ERROR_SUCCESS) {
			// The name on the menu:
			sprintf(buf,"TweakPNG");
			RegSetValueEx(key1,"",0,REG_SZ,(BYTE*)buf,1+strlen(buf));
			RegCloseKey(key1);
		}

		// create HKCR\\[png_auto_file]\shell\tweakpng\\command
		sprintf(buf,"%s\\shell\\tweakpng\\command",logicalname);
		rv=RegCreateKeyEx(HKEY_CLASSES_ROOT,buf,0,NULL,
			REG_OPTION_NON_VOLATILE,KEY_WRITE,NULL,&key1,&disp);
		if(rv==ERROR_SUCCESS) {
			GetModuleFileName(globals.hInst,buf,500);
			sprintf(cmd,"\"%s\" \"%%1\"", buf);
			RegSetValueEx(key1,"",0,REG_SZ,(BYTE*)cmd,1+strlen(cmd));
			RegCloseKey(key1);
		}
	}
}

static void remove_from_explorer_menu(void)
{
	char logicalname[500];
	char buf[500];
	int i;
	int x;


	for(i=0;i<=2;i++) {
		x=get_reg_name_for_ext(fmts[i],logicalname);
		if(x) {
			sprintf(buf,"%s\\shell\\tweakpng\\command",logicalname);
			RegDeleteKey(HKEY_CLASSES_ROOT,buf);
			sprintf(buf,"%s\\shell\\tweakpng",logicalname);
			RegDeleteKey(HKEY_CLASSES_ROOT,buf);
		}
	}
	return;
}

static char *cmprlevel_names[] = { "Default compression",
      "0 = No compression",
	  "1 = Fastest","2","3","4","5","6","7","8",
	  "9 = Best compression",NULL };

static int remember_menusetting;

static BOOL CALLBACK DlgProcPrefs(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	int i;
	WORD id;
	id=LOWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:
		for(i=0;cmprlevel_names[i];i++) {
			SendDlgItemMessage(hwnd,IDC_CMPRLEVEL,CB_ADDSTRING,0,
			(LPARAM)cmprlevel_names[i]);
		}

		if(globals.compression_level == Z_DEFAULT_COMPRESSION) {
			i=0;
		}
		else {
			i=globals.compression_level+1;
		}
		SendDlgItemMessage(hwnd,IDC_CMPRLEVEL,CB_SETCURSEL,(WPARAM)i,0);

		{
			int x;
			x=is_explorer_menu_set();
			CheckDlgButton(hwnd,IDC_ADDTOMENU,x?BST_CHECKED:BST_UNCHECKED);
			remember_menusetting=x;
		}

		return 1;
	case WM_COMMAND:
		switch(id) {
		case IDOK:
			{
				int x;
				x= (IsDlgButtonChecked(hwnd,IDC_ADDTOMENU) == BST_CHECKED);

				if(x!=remember_menusetting) {
					if(x) {
						add_to_explorer_menu();
					}
					else {
						remove_from_explorer_menu();
					}
				}
			}

			i=SendDlgItemMessage(hwnd,IDC_CMPRLEVEL,CB_GETCURSEL,0,0);
			if(i==CB_ERR) i=0;
			if(i==0) globals.compression_level= Z_DEFAULT_COMPRESSION;
			else globals.compression_level= i-1;
			SaveSettings();
			EndDialog(hwnd, 0);
			return 1;

		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
	}
	return 0;
}

static BOOL CALLBACK DlgProcSplitIDAT(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static Chunk *ch=NULL;
	WORD id;
	char buf[500];

	id=LOWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:
		ch = (Chunk*)lParam;
		if(!ch) return 1;
		sprintf(buf,"Size in bytes of first section (0-%d)",ch->length);
		SetDlgItemText(hwnd,IDC_SPLIT_TEXT,buf);

		return 1;

	case WM_DESTROY:
		ch=NULL;
		return 1;

	case WM_COMMAND:
		if(!ch) return 1;
		switch(id) {
		case IDOK:
			{
				int n,rep;

				n=GetDlgItemInt(hwnd,IDC_SPLIT_SIZE,NULL,TRUE);
				if(n<0 || n>(int)ch->length) {
					EndDialog(hwnd, 0);
					return 1;
				}

				rep= (IsDlgButtonChecked(hwnd,IDC_SPLIT_REPEAT)==BST_CHECKED)?1:0;

				if(n==0 && rep) {
					mesg(MSG_E,"Can't divide repeatedly into chunks of zero size.");
					return 1;
				}

				if(ch->parentpng->split_idat(ch->index,n,rep)) {
					EndDialog(hwnd, 0);
				}
				return 1;
			}
		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
	}
	return 0;
}

static BOOL CALLBACK DlgProcSetSig(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static Png *png1=NULL;
	WORD id;
	UINT x;

	id=LOWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:
		png1 = (Png*)lParam;
		if(!png1) return 1;

		switch(png1->imgtype) {
		case IMG_PNG: x=IDC_RADIO1; break;
		case IMG_MNG: x=IDC_RADIO2; break;
		case IMG_JNG: x=IDC_RADIO3; break;
		default: x=0;
		}
		if(x) CheckDlgButton(hwnd,x,BST_CHECKED);

		return 1;

	case WM_DESTROY:
		png1=NULL;
		return 1;

	case WM_COMMAND:
		if(!png1) return 1;
		switch(id) {
		case IDOK:
			{
				int oldtype=png1->imgtype;
				if(IsDlgButtonChecked(hwnd,IDC_RADIO1)==BST_CHECKED) png1->imgtype=IMG_PNG;
				if(IsDlgButtonChecked(hwnd,IDC_RADIO2)==BST_CHECKED) png1->imgtype=IMG_MNG;
				if(IsDlgButtonChecked(hwnd,IDC_RADIO3)==BST_CHECKED) png1->imgtype=IMG_JNG;
				if(png1->imgtype!=oldtype) {
					png1->set_signature();
					png1->modified();
				}
				EndDialog(hwnd, 0);
				return 1;
			}
		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
	}
	return 0;
}



static BOOL CALLBACK DlgProcTools(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id;
	int i;

	static UINT toolsn[] = {IDC_EDIT1,IDC_EDIT2,IDC_EDIT3,IDC_EDIT4,
		IDC_EDIT5,IDC_EDIT6};
	static UINT toolsc[] = {IDC_EDITC1,IDC_EDITC2,IDC_EDITC3,IDC_EDITC4,
		IDC_EDITC5,IDC_EDITC6};
	static UINT toolsp[] = {IDC_EDITP1,IDC_EDITP2,IDC_EDITP3,IDC_EDITP4,
		IDC_EDITP5,IDC_EDITP6};


	id=LOWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:
		for(i=0;i<NUMTOOLS;i++) {
			SetDlgItemText(hwnd,toolsn[i],tools[i].name);
			SetDlgItemText(hwnd,toolsc[i],tools[i].cmd);
			SetDlgItemText(hwnd,toolsp[i],tools[i].params);
			SendDlgItemMessage(hwnd,toolsn[i],EM_LIMITTEXT,MAX_TOOL_NAME  -1,0);
			SendDlgItemMessage(hwnd,toolsc[i],EM_LIMITTEXT,MAX_TOOL_CMD   -1,0);
			SendDlgItemMessage(hwnd,toolsp[i],EM_LIMITTEXT,MAX_TOOL_PARAMS-1,0);
		}
		return 1;

//	case WM_DESTROY:
//		return 1;

	case WM_COMMAND:
		switch(id) {
		case IDOK:
			for(i=0;i<NUMTOOLS;i++) {
				GetDlgItemText(hwnd,toolsn[i],tools[i].name,MAX_TOOL_NAME);
				tools[i].name[MAX_TOOL_NAME-1]='\0';
				GetDlgItemText(hwnd,toolsc[i],tools[i].cmd,MAX_TOOL_CMD);
				tools[i].cmd[MAX_TOOL_CMD-1]='\0';
				GetDlgItemText(hwnd,toolsp[i],tools[i].params,MAX_TOOL_PARAMS);
				tools[i].params[MAX_TOOL_PARAMS-1]='\0';
				SaveSettings();
				EndDialog(hwnd, 0);
			}
			return 1;

		case IDCANCEL:
			EndDialog(hwnd, 0);
			return 1;
		}
	}
	return 0;
}

