// tweakpng.h

#ifndef TWEAKPNG_H
#define TWEAKPNG_H

// comment out the next line to compile without image viewing support
#define TWPNG_SUPPORT_VIEWER

// comment out the next line to compile without zlib
// (required for image viewing, and editing compressed text chunks.)
#define TWPNG_USE_ZLIB


#if defined(TWPNG_USE_ZLIB) || defined(TWPNG_SUPPORT_VIEWER)
#define TWPNG_HAVE_ZLIB
#endif

#define TWEAKPNG_VER_STRING      "1.2.1"
#define TWEAKPNG_COPYRIGHT_DATE  "1999-2004"
#define TWEAKPNG_HOMEPAGE        "http://pobox.com/~jason1/tweakpng/"


#define ID_STBAR      19000
//#define ID_IMGVIEWER     19100
#define ID_UPDATEVIEWER  19101

#define IMG_UNKNOWN 0
#define IMG_PNG 1
#define IMG_MNG 2
#define IMG_JNG 3


#define CHUNK_UNKNOWN    0
#define CHUNK_IHDR  1
#define CHUNK_IEND  2
#define CHUNK_IDAT  3
#define CHUNK_PLTE  4
#define CHUNK_bKGD  5
#define CHUNK_cHRM  6
#define CHUNK_gAMA  7
#define CHUNK_hIST  8
#define CHUNK_pHYs  9
#define CHUNK_sBIT  10
#define CHUNK_tEXt  11
#define CHUNK_tIME  12
#define CHUNK_tRNS  13
#define CHUNK_zTXt  14

// v1.2 chunks:
#define CHUNK_sRGB  15
#define CHUNK_iCCP  16
#define CHUNK_iTXt  17
#define CHUNK_sPLT  18

// MNG-specific chunks
#define CHUNK_MHDR 200
#define CHUNK_LOOP 201
#define CHUNK_ENDL 202
#define CHUNK_DEFI 203
#define CHUNK_JHDR 204
#define CHUNK_BASI 205
#define CHUNK_CLON 206
#define CHUNK_DHDR 207
#define CHUNK_PAST 208
#define CHUNK_DISC 209
#define CHUNK_BACK 210
#define CHUNK_FRAM 211
#define CHUNK_MOVE 212
#define CHUNK_CLIP 213
#define CHUNK_SHOW 214
#define CHUNK_TERM 215
#define CHUNK_SAVE 216
#define CHUNK_SEEK 217
#define CHUNK_MEND 218

#define CHUNK_eXPI 230
#define CHUNK_fPRI 231
#define CHUNK_nEED 232
#define CHUNK_pHYg 233
#define CHUNK_JDAT 234
#define CHUNK_JSEP 235
#define CHUNK_PROM 236
#define CHUNK_IPNG 237
#define CHUNK_PPLT 238
#define CHUNK_IJNG 239
#define CHUNK_DROP 240
#define CHUNK_DBYK 241
#define CHUNK_ORDR 242


// registered PNG extensions:
#define CHUNK_oFFs  501
#define CHUNK_pCAL  502
#define CHUNK_sCAL  503
#define CHUNK_gIFg  504
#define CHUNK_gIFx  505
#define CHUNK_gIFt  506   /* deprecated */
#define CHUNK_fRAc  507


#define MSG_S 1 // severe
#define MSG_E 0 // error
#define MSG_W 2 // warning
#define MSG_I 3 // information


#define CRCCOMPL(c) ((c)^0xffffffff)
#define CRCINIT (CRCCOMPL(0))


struct windowpos_struct {
	int x;
	int y;
	int w;
	int h;
	int max;
};


struct windowprefs_struct {
	// don't change or rearrange the existing members of this
	// structure, or saved settings will be screwed up
	int version;
	DWORD column_width[5];
	struct windowpos_struct main;
	struct windowpos_struct text;
	struct windowpos_struct plte;
	struct windowpos_struct viewer;
};

struct globals_struct {
	HINSTANCE hInst;
	HWND hwndMain;
	int compression_level;
	int use_gamma;
	COLORREF custombgcolor;
	int use_custombg;
	int use_imagebg;
	int vborder; // viewer border width

#define TWPNG_VS_FIT 1
#define TWPNG_VS_1_8 2
#define TWPNG_VS_1_4 3
#define TWPNG_VS_1_2 4
#define TWPNG_VS_1   5
#define TWPNG_VS_2   6
#define TWPNG_VS_4   7
#define TWPNG_VS_8   8
	int vsize;

#define TWPNG_WBG_SAMEASIMAGE 0
#define TWPNG_WBG_SYSDEFAULT  1
#define TWPNG_WBG_CUSTOM      2
	int window_bgcolor;

	COLORREF custcolors[16];
	int autoopen_viewer;
	HCURSOR hcurDrag,hcurDrag2;
	int viewer_imgpos_x, viewer_imgpos_y;
};

#define MAX_TOOL_NAME   80
#define MAX_TOOL_CMD    128
#define MAX_TOOL_PARAMS 80

typedef struct tools_t_struct {
	char name[MAX_TOOL_NAME];
	char cmd[MAX_TOOL_CMD];
	char params[MAX_TOOL_PARAMS];
} tools_t;



DWORD update_crc(DWORD crc, unsigned char *buf, int len);
void write_int32(unsigned char *buf, DWORD x);
DWORD read_int32(unsigned char *x);
int read_int16(unsigned char *x);
void write_int16(unsigned char *buf, int x);
int GetLVSelection(HWND hwnd);
void SetLVSelection(HWND hwnd, int pos, int num);
int get_name_from_id(char *name, int x);
void mesg(int severity, char *fmt, ...);
int choose_color_dialog(HWND hwnd, unsigned char *redp,
						unsigned char *greenp, unsigned char *bluep);
void DroppedFiles(HDROP hDrop);

LRESULT CALLBACK WndProcEditPal(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

int StoreWindowPos(HWND hwnd, struct windowpos_struct *q);
void update_viewer();

class Png;

class Chunk {
public:
	Chunk(void);
	~Chunk(void);

	void after_init(void);

	int write_to_file(HANDLE fh, int exp);
	void get_text_descr(char *buf);    // buf must be 1000 chars or more
	void get_text_descr_generic(char *buf);    // buf must be 1000 chars or more
	int get_chunk_type_id(void); // convert type into an integer
	int copy_to_memory(unsigned char *m);
	int init_from_memory(unsigned char *buf, int);

	DWORD calc_crc(void);  // calculates CRC, does not modify it

	int edit(void);  // generic edit; calls the right edit_*() function

	void chunkmodified(void);

	int is_critical(void);
	int is_public(void);
	int is_safe_to_copy(void);

	DWORD length;     /* length of the DATA field */
	char chunktype[5];
	int chunktype_id;
	DWORD crc;
	unsigned char *data;
	Png *parentpng;
	int get_text_info(unsigned char **keywordp);
	int set_text_info(int chunktype_id,unsigned char *keyword,unsigned char *indata);

	int index; // sometimes used to hold the position in the parent chunk array
	int flag;  // sometimes used to remember if a chunk was selected

	// used in handling text chunks
	unsigned char *uncompressed_data;   // used only for zTXt
	int uncompressed_data_size;
	int uncompressed_data_tried;  // did we try and fail to uncompress?

private:

	void describe_IHDR(char *buf);
	void describe_MHDR(char *buf);
	void describe_IEND(char *buf);
	void describe_bKGD(char *buf);
	void describe_PLTE(char *buf);
	void describe_tRNS(char *buf);
	void describe_gAMA(char *buf);
	void describe_sBIT(char *buf);
	void describe_cHRM(char *buf);
	void describe_pHYs(char *buf);
	void describe_tIME(char *buf);
	void describe_hIST(char *buf);
	void describe_text(char *buf, int id);
	void describe_sRGB(char *buf);
	void describe_keyword_chunk(char *buf, int ct);

	int edit_plte_info(void);

};


class Png {

public:
	Png(char *fn);
	Png(void);

	~Png(void);

	int valid;

	int write_file(char *fn);
	int write_to_mem(unsigned char **pmem, int *plen);
	void fill_listbox(HWND hwnd);
	void delete_chunk(int);
	void move_chunk(int,int);

	void updatestbar(void);
	void modified(void); // only call if something really changed. also calls updatestbar
	
	void set_signature(void);
	int check_validity(int msgmode);
	void edit_chunk(int);
	int split_idat(int cn, int size, int repeat);
	void insert_chunks(int pos, int num, int init);
	void new_chunk(int chunktype_id);
	Chunk *find_first_chunk(int chunktype_id, int *index);
	

	int imgtype;
	int num_chunks;
	DWORD pngfilesize;   // used when loading from disk

	//info from the header
	DWORD width;
	DWORD height;
	unsigned char bitdepth;
	unsigned char colortype;
	unsigned char compression;
	unsigned char compressionfilter;
	unsigned char interlace;

	int create_display_window(void);

	Chunk **chunk;

	char filename[MAX_PATH];
	int named;
	int dirty; // has file been modified?

private:
	int chunks_alloc;    /* alloc'd length of the chunk array */
	void update_row(HWND hwnd,int n);

	unsigned char signature[8];

	void init_new_chunk(int);

	int read_signature(HANDLE fh);
	int read_next_chunk(HANDLE fh);

};


class Viewer {
public:
	Viewer(HWND parent);
	~Viewer();

	static void GlobalInit();
	void Close();
	void Update(void *mappedpng, int msize, int imgtype);
	HWND hwndViewer;

private:
	LPBITMAPINFOHEADER m_dib;
	void *m_bits;

	int m_imghasbgcolor;
	COLORREF m_imgbgcolor;

	int m_errorflag; // 0 = no error
	char m_errormsg[200];

	static LRESULT CALLBACK WndProcViewer(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
	void FreeImage();
	int m_adjwidth, m_adjheight; // after adjusting for non-square pixels
	int m_stretchedwidth, m_stretchedheight;
	int m_imgpos_x, m_imgpos_y;
	int m_dragging;
	int m_dragstart_x, m_dragstart_y;
	void GoodScrollPos();
	void CalcStretchedSize();
	RECT m_clientrect;
};


#endif
