// tweakpng image viewer

#include <windows.h>
#include "tweakpng.h"
#include <pngdib.h>

#include "resource.h"

#ifdef TWPNG_SUPPORT_VIEWER

extern Viewer *g_viewer;
extern struct windowprefs_struct window_prefs;
extern struct globals_struct globals;

void Viewer::GlobalInit()
{
	WNDCLASS wc;

	ZeroMemory(&wc,sizeof(WNDCLASS));
	wc.style = CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS;
	wc.lpfnWndProc = (WNDPROC)WndProcViewer;
	wc.hInstance = globals.hInst;
	wc.hIcon = LoadIcon(globals.hInst,"ICONMAIN");
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = NULL; //(HBRUSH)GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName =  "MENUVIEWER";
	wc.lpszClassName = "TPNGVIEWER";
	RegisterClass(&wc);

	globals.viewer_imgpos_x=globals.vborder;
	globals.viewer_imgpos_y=globals.vborder;
}


// the constructor creates the window
Viewer::Viewer(HWND parent)
{
	DWORD styles;
	hwndViewer=NULL;
	m_dib = NULL;
	m_bits = NULL;
	m_errorflag=0;

	styles = WS_OVERLAPPEDWINDOW;
	styles &= ~WS_MINIMIZEBOX;
	hwndViewer=CreateWindow(
		"TPNGVIEWER","TweakPNG Image Viewer",styles,
		window_prefs.viewer.x,  // horizontal position
		window_prefs.viewer.y,	// vertical position
		window_prefs.viewer.w,	// width
		window_prefs.viewer.h,  // height
		parent,NULL,globals.hInst,NULL);
	if(!hwndViewer) return;
	ShowWindow(hwndViewer,window_prefs.viewer.max?SW_SHOWMAXIMIZED:SW_SHOWNOACTIVATE);
	GetClientRect(hwndViewer,&m_clientrect);

	//It isn't safe to allow drag and drop into the viewer window,
	//because it will drag and drop will still be available even
	//if the main window has a dialog box window open. In that case,
	//there can be big problems when the dialog is closed and the
	//program tries to update a PNG file that has suddenly and
	//unexpectedly been replace with a different one.
	//I haven't been able to find an easy workaround to this.
	//DragAcceptFiles(hwndViewer,TRUE);

	m_dragging=0;
	m_imgpos_x = globals.viewer_imgpos_x;
	m_imgpos_y = globals.viewer_imgpos_y;
}

Viewer::~Viewer()
{
	FreeImage();
	globals.viewer_imgpos_x = m_imgpos_x;
	globals.viewer_imgpos_y = m_imgpos_y;
}

// sets m_strechedwidth based on m_adjwidth and globals.vsize, etc.
void Viewer::CalcStretchedSize()
{
	if(!m_dib) return;
	switch(globals.vsize) {
	case TWPNG_VS_FIT:
		{
			int xavail, yavail;
			double ratio1,ratio2;
			xavail=m_clientrect.right-globals.vborder*2;
			yavail=m_clientrect.bottom-globals.vborder*2;
			if(xavail<1) xavail=1;
			if(yavail<1) yavail=1;
			ratio1 = (double)yavail/(double)xavail;
			ratio2 = (double)m_adjheight/(double)m_adjwidth;
			if(ratio1>ratio2) {
				m_stretchedwidth = xavail;
				m_stretchedheight = (int)(0.5+((double)m_adjheight)*((double)xavail)/(double)m_adjwidth);
			}
			else {
				m_stretchedwidth = (int)(0.5+((double)m_adjwidth)*((double)yavail)/(double)m_adjheight);
				m_stretchedheight = yavail;
			}
		}
		break;
	case TWPNG_VS_1_8:
		m_stretchedwidth = (int)(0.5+((double)m_adjwidth)/8.0);
		m_stretchedheight = (int)(0.5+((double)m_adjheight)/8.0);
		break;
	case TWPNG_VS_1_4:
		m_stretchedwidth = (int)(0.5+((double)m_adjwidth)/4.0);
		m_stretchedheight = (int)(0.5+((double)m_adjheight)/4.0);
		break;
	case TWPNG_VS_1_2:
		m_stretchedwidth = (1+m_adjwidth)/2;
		m_stretchedheight = (1+m_adjheight)/2;
		break;
	case TWPNG_VS_2:
		m_stretchedwidth = m_adjwidth*2;
		m_stretchedheight = m_adjheight*2;
		break;
	case TWPNG_VS_4:
		m_stretchedwidth = m_adjwidth*4;
		m_stretchedheight = m_adjheight*4;
		break;
	case TWPNG_VS_8:
		m_stretchedwidth = m_adjwidth*8;
		m_stretchedheight = m_adjheight*8;
		break;
	default:
		m_stretchedwidth = m_adjwidth;
		m_stretchedheight = m_adjheight;
	}
	if(m_stretchedwidth<1) m_stretchedwidth=1;
	if(m_stretchedheight<1) m_stretchedheight=1;

}


// process image
void Viewer::Update(void *mappedpng, int msize, int imgtype)
{
	PNGDIB *p2d;
	int rv;
	int dens_x, dens_y, dens_units;
	unsigned char tmpr,tmpg,tmpb;

	m_errorflag=0;
	p2d=NULL;
	FreeImage();

	if(!mappedpng) goto abort;

	if(imgtype==IMG_MNG) {
		m_errorflag=1;
		lstrcpy(m_errormsg,"Viewer doesn't support MNG files.");
		goto abort;
	}
	else if(imgtype==IMG_JNG) {
		m_errorflag=1;
		lstrcpy(m_errormsg,"Viewer doesn't support JNG files.");
		goto abort;
	}

	p2d = pngdib_p2d_init();

	pngdib_p2d_set_png_memblk(p2d,mappedpng,msize);

	pngdib_p2d_set_gamma_correction(p2d, globals.use_gamma?1:0, PNGDIB_DEFAULT_SCREEN_GAMMA);

	pngdib_p2d_set_use_file_bg(p2d,globals.use_imagebg?1:0);
	if(globals.use_custombg) {
		pngdib_p2d_set_custom_bg(p2d,GetRValue(globals.custombgcolor),
			GetGValue(globals.custombgcolor),GetBValue(globals.custombgcolor));
	}

	rv=pngdib_p2d_run(p2d);
	if(rv!=PNGD_E_SUCCESS) {
		lstrcpyn(m_errormsg,pngdib_get_error_msg(p2d),200);
		m_errorflag=1;
		goto abort;
	}
	rv=pngdib_p2d_get_dib(p2d,&m_dib,NULL);
	rv=pngdib_p2d_get_dibbits(p2d, &m_bits, NULL, NULL);

	m_adjwidth = m_dib->biWidth;
	m_adjheight = m_dib->biHeight;
	rv=pngdib_p2d_get_density(p2d, &dens_x, &dens_y, &dens_units);
	if(rv) {
		if(dens_x!=dens_y && dens_x>0 && dens_y>0 &&
			10*dens_x>dens_y && 10*dens_y>dens_x)
		{
			if(dens_x>dens_y) {
				m_adjheight = (int) (0.5+ (double)m_adjheight * ((double)dens_x/(double)dens_y) );
			}
			else {
				m_adjwidth = (int) (0.5+ (double)m_adjwidth * ((double)dens_y/(double)dens_x) );
			}
		}
	}

	CalcStretchedSize();

	m_imghasbgcolor=0;
	if(pngdib_p2d_get_bgcolor(p2d,&tmpr,&tmpg,&tmpb)) {
		m_imghasbgcolor=1;
		m_imgbgcolor = RGB(tmpr,tmpg,tmpb);
	}

abort:
	if(p2d) pngdib_done(p2d);

	if(hwndViewer) {
		InvalidateRect(hwndViewer,NULL,TRUE);
	}
}

void Viewer::FreeImage()
{
	if(m_dib) {
		pngdib_p2d_free_dib(NULL,m_dib);
		m_dib=NULL;
		m_bits=NULL;
		m_errorflag=0;
	}
}

void Viewer::Close()
{
	if(hwndViewer) {
		DestroyWindow(hwndViewer);
	}
}

// make sure scroll position is within acceptable limits
void Viewer::GoodScrollPos()
{
	int b;
	b=globals.vborder;

	if(m_stretchedwidth+2*b <= m_clientrect.right) {
		//imgpos must be between 0 and (windowwidth-imgwidth)
		if(m_imgpos_x<b) m_imgpos_x=b;
		if(m_imgpos_x > (m_clientrect.right-m_stretchedwidth-b))
			m_imgpos_x = m_clientrect.right-m_stretchedwidth-b;
	}
	else {
		// must be between -(imgwidth-windowwidth) and 0
		if(m_imgpos_x < (m_clientrect.right-m_stretchedwidth-b))
			m_imgpos_x = m_clientrect.right-m_stretchedwidth-b;
		if(m_imgpos_x>b) m_imgpos_x=b;
	}

	if(m_stretchedheight+2*b <= m_clientrect.bottom) {
		//imgpos must be between 0 and (windowwidth-imgwidth)
		if(m_imgpos_y<b) m_imgpos_y=b;
		if(m_imgpos_y > (m_clientrect.bottom-m_stretchedheight-b))
			m_imgpos_y = m_clientrect.bottom-m_stretchedheight-b;
	}
	else {
		// must be between -(imgwidth-windowwidth) and 0
		if(m_imgpos_y < (m_clientrect.bottom-m_stretchedheight-b))
			m_imgpos_y = m_clientrect.bottom-m_stretchedheight-b;
		if(m_imgpos_y>0) m_imgpos_y=b;
	}
}


LRESULT CALLBACK Viewer::WndProcViewer(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id;
	POINTS pts;
	Viewer *v;

	if(!g_viewer) goto exit1;
	v=g_viewer;

	switch(msg) {
	case WM_ERASEBKGND:
		return 1;
	case WM_SETCURSOR:
		if(LOWORD(lParam)==HTCLIENT) {
			if(v->m_dib) {
				SetCursor(globals.hcurDrag);
				return 0;
			}
		}
		break;
	case WM_PAINT:
		{
			PAINTSTRUCT paintStruct;
			HDC hdc;
			HBRUSH hbr1;
			int rv;

			hdc = BeginPaint(hwnd, &paintStruct);

			if(v->m_dib) {

				v->GoodScrollPos();

				// paint the image
				SetStretchBltMode(hdc,COLORONCOLOR);
				rv=StretchDIBits(hdc,
					v->m_imgpos_x,v->m_imgpos_y,v->m_stretchedwidth,v->m_stretchedheight,
					0,0,v->m_dib->biWidth,v->m_dib->biHeight,
					v->m_bits,(LPBITMAPINFO)v->m_dib,
					DIB_RGB_COLORS,SRCCOPY);
				if(rv==GDI_ERROR || rv<1) {
					RECT r1;
					// Failure probably means we've exceeded the video system's
					// stretching limits. Draw a black rectangle instead.
					r1.left=v->m_imgpos_x; r1.top=v->m_imgpos_y;
					r1.right=v->m_imgpos_x+v->m_stretchedwidth;
					r1.bottom=v->m_imgpos_y+v->m_stretchedheight;
					FillRect(hdc,&r1,(HBRUSH)GetStockObject(BLACK_BRUSH));
				}

				// paint the background
				ExcludeClipRect(hdc,
					v->m_imgpos_x, v->m_imgpos_y,
					v->m_imgpos_x+v->m_stretchedwidth, v->m_imgpos_y+v->m_stretchedheight);

				if(globals.window_bgcolor==TWPNG_WBG_SYSDEFAULT) {
					FillRect(hdc,&v->m_clientrect,GetSysColorBrush(COLOR_WINDOW));
				}
				else {
					
					if(globals.window_bgcolor==TWPNG_WBG_SAMEASIMAGE && v->m_imghasbgcolor)
						hbr1 = CreateSolidBrush(v->m_imgbgcolor);
					else
						hbr1 = CreateSolidBrush(globals.custombgcolor);
					FillRect(hdc,&v->m_clientrect,hbr1);
					DeleteObject(hbr1);
				}
			}
			else {
				FillRect(hdc,&v->m_clientrect,GetSysColorBrush(COLOR_WINDOW));

				if(v->m_errorflag) {
					SetTextColor(hdc,RGB(128,0,0));
					SetBkColor(hdc,RGB(255,255,255));
					SetBkMode(hdc,OPAQUE);
					SelectObject(hdc,GetStockObject(ANSI_VAR_FONT));
					DrawText(hdc,v->m_errormsg,-1,&v->m_clientrect,
						DT_NOPREFIX|DT_WORDBREAK);
				}
			}

			EndPaint(hwnd, &paintStruct);
			return 0;
		}

	case WM_INITMENU:
		{
			HMENU m;
			m=(HMENU)wParam;
			CheckMenuItem(m,ID_GAMMACORRECT,MF_BYCOMMAND|
				(globals.use_gamma?MF_CHECKED:MF_UNCHECKED));
			CheckMenuItem(m,ID_BG_BKGD,MF_BYCOMMAND|
				(globals.use_imagebg?MF_CHECKED:MF_UNCHECKED));
			CheckMenuItem(m,ID_BG_CUSTOM,MF_BYCOMMAND|
				(globals.use_custombg?MF_CHECKED:MF_UNCHECKED));

			CheckMenuItem(m,ID_WBG_SAMEASIMAGE,MF_BYCOMMAND|
				(globals.window_bgcolor==TWPNG_WBG_SAMEASIMAGE?MF_CHECKED:MF_UNCHECKED));
			CheckMenuItem(m,ID_WBG_SYSDEFAULT,MF_BYCOMMAND|
				(globals.window_bgcolor==TWPNG_WBG_SYSDEFAULT?MF_CHECKED:MF_UNCHECKED));
			CheckMenuItem(m,ID_WBG_CUSTOM,MF_BYCOMMAND|
				(globals.window_bgcolor==TWPNG_WBG_CUSTOM?MF_CHECKED:MF_UNCHECKED));
			CheckMenuItem(m,ID_ZOOM_FIT,MF_BYCOMMAND|(globals.vsize==TWPNG_VS_FIT?MF_CHECKED:MF_UNCHECKED));
			CheckMenuItem(m,ID_ZOOM_1_8,MF_BYCOMMAND|(globals.vsize==TWPNG_VS_1_8?MF_CHECKED:MF_UNCHECKED));
			CheckMenuItem(m,ID_ZOOM_1_4,MF_BYCOMMAND|(globals.vsize==TWPNG_VS_1_4?MF_CHECKED:MF_UNCHECKED));
			CheckMenuItem(m,ID_ZOOM_1_2,MF_BYCOMMAND|(globals.vsize==TWPNG_VS_1_2?MF_CHECKED:MF_UNCHECKED));
			CheckMenuItem(m,ID_ZOOM_1,MF_BYCOMMAND|(globals.vsize==TWPNG_VS_1?MF_CHECKED:MF_UNCHECKED));
			CheckMenuItem(m,ID_ZOOM_2,MF_BYCOMMAND|(globals.vsize==TWPNG_VS_2?MF_CHECKED:MF_UNCHECKED));
			CheckMenuItem(m,ID_ZOOM_4,MF_BYCOMMAND|(globals.vsize==TWPNG_VS_4?MF_CHECKED:MF_UNCHECKED));
			//CheckMenuItem(m,ID_ZOOM_8,MF_BYCOMMAND|(globals.vsize==TWPNG_VS_8?MF_CHECKED:MF_UNCHECKED));
			return 0;
		}

	case WM_DESTROY:
		StoreWindowPos(hwnd,&window_prefs.viewer);
		delete g_viewer;
		g_viewer=NULL;
		SetFocus(globals.hwndMain);
		return 0;
		
	case WM_SIZE:
		GetClientRect(hwnd,&v->m_clientrect);
		if(globals.vsize==TWPNG_VS_FIT) v->CalcStretchedSize();
		return 0;

	case WM_DROPFILES:
		DroppedFiles((HDROP)wParam);
		return 0;

	case WM_LBUTTONDBLCLK:
		if(globals.vsize==TWPNG_VS_1 || globals.vsize==TWPNG_VS_FIT) {
			if(globals.vsize==TWPNG_VS_1)
				globals.vsize=TWPNG_VS_FIT;
			else
				globals.vsize=TWPNG_VS_1;
			v->CalcStretchedSize();
			InvalidateRect(hwnd,NULL,0);
		}
		return 0;

	case WM_LBUTTONDOWN:
		if(v->m_dragging) return 0;
		if(!v->m_dib) return 0;
		SetCapture(hwnd);
		SetCursor(globals.hcurDrag2);
		pts = MAKEPOINTS(lParam);
		v->m_dragstart_x = v->m_imgpos_x - pts.x;
		v->m_dragstart_y = v->m_imgpos_y - pts.y;
		v->m_dragging = 1;
		return 0;

	case WM_LBUTTONUP:
		SetCapture(NULL);
		v->m_dragging=0;
		return 0;


	case WM_MOUSEMOVE:
		if(v->m_dragging) {
			pts = MAKEPOINTS(lParam);
			v->m_imgpos_x = v->m_dragstart_x + pts.x;
			v->m_imgpos_y = v->m_dragstart_y + pts.y;
			InvalidateRect(hwnd,NULL,TRUE);
		}
		return 0;

	case WM_CLOSE: // The user clicked the [x] for the window.
		globals.autoopen_viewer=0;
		break;  // DefWindowProc will send a WM_DESTROY message

	case WM_COMMAND:
		id=LOWORD(wParam);
		switch(id) {
		case ID_CLOSE:
			globals.autoopen_viewer=0;
			DestroyWindow(hwnd);
			return 0;
		case ID_GAMMACORRECT:
			globals.use_gamma = !globals.use_gamma;
			update_viewer();
			return 0;
		case ID_BG_CUSTOM:
			globals.use_custombg = !globals.use_custombg;
			update_viewer();
			return 0;
		case ID_BG_BKGD:
			globals.use_imagebg = !globals.use_imagebg;
			update_viewer();
			return 0;
		case ID_WBG_SAMEASIMAGE:
			globals.window_bgcolor = TWPNG_WBG_SAMEASIMAGE;
			InvalidateRect(hwnd,NULL,0);
			break;
		case ID_WBG_SYSDEFAULT:
			globals.window_bgcolor = TWPNG_WBG_SYSDEFAULT;
			InvalidateRect(hwnd,NULL,0);
			break;
		case ID_WBG_CUSTOM:
			globals.window_bgcolor = TWPNG_WBG_CUSTOM;
			InvalidateRect(hwnd,NULL,0);
			break;

		case ID_SETBG:
			{
				unsigned char tmpr,tmpg,tmpb;
				tmpr=GetRValue(globals.custombgcolor);
				tmpg=GetGValue(globals.custombgcolor);
				tmpb=GetBValue(globals.custombgcolor);
				if(choose_color_dialog(hwnd,&tmpr,&tmpg,&tmpb)) {
					globals.custombgcolor = RGB(tmpr,tmpg,tmpb);
					update_viewer();
				}
				return 0;
			}

		case ID_ZOOM_FIT: globals.vsize=TWPNG_VS_FIT; goto rezoom;
		case ID_ZOOM_1_8: globals.vsize=TWPNG_VS_1_8; goto rezoom;
		case ID_ZOOM_1_4: globals.vsize=TWPNG_VS_1_4; goto rezoom;
		case ID_ZOOM_1_2: globals.vsize=TWPNG_VS_1_2; goto rezoom;
		case ID_ZOOM_2: globals.vsize=TWPNG_VS_2; goto rezoom;
		case ID_ZOOM_4: globals.vsize=TWPNG_VS_4; goto rezoom;
		//case ID_ZOOM_8: globals.vsize=TWPNG_VS_8; goto rezoom;
		case ID_ZOOM_1: globals.vsize=TWPNG_VS_1;
rezoom:
			v->CalcStretchedSize();
			InvalidateRect(hwnd,NULL,0);
			break;

		}
	}

exit1:
	return DefWindowProc(hwnd, msg, wParam, lParam);

}

#endif TWPNG_SUPPORT_VIEWER
